# File: kernels.py
# File created: Apr 13, 2017 
# Author: Yinhao Zhu (yzhu10@nd.edu)

"""
Implementation of kernels for GP
"""


from __future__ import absolute_import, print_function

from torch.autograd import Variable
import torch as th
import numpy as np

from .util import as_variable, TensorType, squared_distance
from .model import Model, Param


def _k_shape(X, X2):
    """
    Shape of a kernel with these inputs
    :param X:
    :param X2:
    :return:
    """
    return (X.size(0),) * 2 if X2 is None else (X.size(0), X2.size(0))


class Kernel(Model):
    """
    Base class for kernels
    """
    def __init__(self, input_dim):
        self.input_dim = int(input_dim)
        # self.name_kernel = name
        super(Kernel, self).__init__()

    def __add__(self, other):
        return Sum(self, other)

    def __mul__(self, other):
        return Product(self, other)

    def _validate_ard_shape(self, x, ARD=None):
        """
        Validates the shape of a potentially ARD hyperparameter

        :param name: The name of the parameter (used for error messages)
        :param x: A scalar or an array.
        :param ARD: None, False, or True. If None, infers ARD from value.
        :return: Tuple (value, ARD),
            val is a 1D np.ndarray
            ARD is a bool
        """
        if ARD is None:
            ARD = np.asarray(x).squeeze().shape != ()

        x = x * np.ones(self.input_dim)
        correct_shape = (self.input_dim,)

        if x.shape != correct_shape:
            raise ValueError(
                "shape of possibly-ARD param does not match input_dim")

        return x, ARD


class Static(Kernel):
    """
    Kernels that don't depend on the value of the inputs are 'Static'.  The only
    parameter is a variance.
    """

    def __init__(self, input_dim, variance=1.0):
        super().__init__(input_dim)
        self.variance = Param(TensorType([variance]), requires_transform=True)

    def Kdiag(self, X):
        return self.variance.transform().expand(X.size(0))


class White(Static):
    """
    The White kernel
    """

    def K(self, X, X2=None, presliced=False):
        if X2 is None:
            return self.variance.transform().expand(X.size(0)).diag()
        else:
            return Variable(th.zeros(*_k_shape(X, X2)).type(TensorType))


class Constant(Static):
    """
    The Constant (aka Bias) kernel
    """

    def K(self, X, X2=None, presliced=False):
        return self.variance.transform().expand(*_k_shape(X, X2))


class Bias(Constant):
    pass


class Stationary(Kernel):
    """
    Base class for stationary kernels, which only depend on r = || x - x'||
    This class handles 'ARD' behaviour, which stands for 'Automatic Relevance
    Determination'. This means that the kernel has one lengthscale per
    dimension, otherwise the kernel is isotropic (has a single lengthscale).
    """

    def __init__(self, input_dim, variance=1.0, length_scales=None, ARD=False):
        """
        Args:
            input_dim (int): the dimension of the input
            variance (float): initial value for the signal variance
            length_scales (np.ndarray): initial value for length scale
                defaults to 1.0 (ARD=False) or np.ones(input_dim) (ARD=True)
            ARD (bool): ARD specifies whether the kernel has one length scale
                per dimension (ARD=True) or a single length scale (ARD=False).
        """
        super(Stationary, self).__init__(input_dim)
        # self.variance_param = Parameter(th.FloatTensor(SoftplusInv(variance)))
        self.variance = Param(TensorType([variance]), requires_transform=True)
        self.ARD = ARD
        if ARD:
            if length_scales is None:
                length_scales = np.ones(input_dim)
            elif isinstance(length_scales, np.ndarray):
                assert len(length_scales) == input_dim
            else:
                length_scales = length_scales * np.ones(input_dim)

            self.length_scales = \
                Param(TensorType(length_scales), requires_transform=True)
        else:
            if length_scales is None:
                length_scales = 1.0
            self.length_scales = \
                Param(TensorType([length_scales]), requires_transform=True)

    def squared_dist(self, X, X2):
        """
        Returns the SCALED squared distance between X and X2.
        """
        return squared_distance(X / self.length_scales.transform()) \
            if X2 is None else \
            squared_distance(X / self.length_scales.transform(), 
                         X2 / self.length_scales.transform())

    def dist(self, X, X2):
        """
        Matrix of Euclidean distances between points

        Args:
            X (Variable): Matrix of vectors
            X2 (Variable): (Optional) matrix of vectors.  If None, X2 = X.
        Returns:
            (Variable) entry (i, j) is the distance between X[i, :] and X2[i, :]
        """
        # Clamp to slightly positive so that the gradient of sqrt(x) is finite
        return th.sqrt(th.clamp(self.squared_dist(X, X2), min=1e-40))

    def Kdiag(self, X):
        if isinstance(X, np.ndarray):
            # input is a data matrix; each row represents one instance
            X = as_variable(X)
        # return a vector
        return self.variance.transform().expand(X.size(0))


class Exp(Stationary):
    """
    Exponential Kernel

    k(x, y; variance, length_scale) = variance * exp(-|x-y| / length_scale)
    """
    def K(self, X, X2=None):
        return self.variance.transform() * th.exp(-self.dist(X, X2))


class Matern12(Exp):
    pass


class Matern32(Stationary):
    def K(self, X, X2=None):
        r3 = Variable(TensorType([np.sqrt(3.0)])) * self.dist(X, X2)
        return self.variance.transform() * (1.0 + r3) * th.exp(-r3)


class Matern52(Stationary):
    def K(self, X, X2=None):
        r = self.dist(X, X2)
        s5 = Variable(TensorType([np.sqrt(5.0)]))
        return self.variance.transform() * \
               (1.0 + s5 * r + 5.0 / 3.0 * r * r) * \
               th.exp(-s5 * r)


class Rbf(Stationary):
    """
    The Radial Basis Function (RBF) or Squared Exponential / Gaussian Kernel
    """
    def K(self, X, X2=None):
        r2 = self.squared_dist(X, X2)
        return self.variance.transform() * th.exp(-r2 / 2.0)


SquaredExponential = Rbf


class Linear(Kernel):
    """
    The linear kernel
    """

    def __init__(self, input_dim, variance=1.0, ARD=None):
        """
        - input_dim is the dimension of the input to the kernel
        - variance is the (initial) value for the variance parameter(s)
          if ARD=True, there is one variance per input
        - active_dims is a list of length input_dim which controls
          which columns of X are used.
        """
        super().__init__(input_dim)

        variance, self.ARD = self._validate_ard_shape(variance, ARD)
        self.variance = Param(TensorType(variance), requires_transform=True)

    def K(self, X, X2=None):
        if X2 is None:
            return th.mm(X * self.variance.transform(), X.t())
        else:
            return th.mm(X * self.variance.transform(), X2.t())

    def Kdiag(self, X):
        return th.sum(X * X * self.variance.transform(), 1)


class Combination(Kernel):
    """
    A combination of two kernels (e.g. Sum or Product)

    This will later get extended to arbitrary numbers of kernels, but we'll keep
    it to pairs at the moment, and you can use Combinations of Combinations to
    build more expressive kernels.
    """
    def __init__(self, kern1, kern2):
        if not kern1.input_dim == kern2.input_dim:
            raise ValueError("Kernels need the same input_dim")
        super().__init__(input_dim=kern1.input_dim)

        self.kern1 = kern1
        self.kern2 = kern2


class Product(Combination):
    """
    Product kernel

    CAREFUL: don't leave kern1 and kern2 with trainable variances unless they
    have priors!
    """
    def K(self, X, X2=None):
        return self.kern1.K(X, X2) * self.kern2.K(X, X2)

    def Kdiag(self, X):
        return self.kern1.Kdiag(X) * self.kern2.Kdiag(X)


class Sum(Combination):
    def K(self, X, X2=None):
        return self.kern1.K(X, X2) + self.kern2.K(X, X2)

    def Kdiag(self, X):
        return self.kern1.Kdiag(X) + self.kern2.Kdiag(X)



# File: test_likelihoods.py
# File Created: Wednesday, 13th February 2019 9:33:40 pm
# Author: Steven Atkinson (steven@atkinson.mn)

"""
Tests for likelihood classes
"""

import torch as th
from torch.autograd import Variable
import pytest

from gptorch import likelihoods
import gptorch.model
from gptorch.util import TensorType


class TestGaussian(object):
    """
    Tests for Gaussian likelihood
    """

    def test_init(self):
        """
        Ensure that initialization succeeds
        """
        likelihoods.Gaussian()
        self._standard_likelihood()

    def test_variance(self):
        """
        Test variance API
        """
        lik = self._standard_likelihood()
        
        # Type
        assert isinstance(lik.variance, gptorch.model.Param)

        # Value
        print(type(lik.variance.transform()))
        assert lik.variance.transform().data.numpy() == \
            pytest.approx(self._expected_likelihood_variance)

    def test_logp(self):
        """
        Log-density
        """
        lik = self._standard_likelihood()
        mean = Variable(TensorType([0.0]))
        target = Variable(TensorType([0.1]))
        expected_logp = 0.8836465597893728

        # API
        logp = lik.logp(mean, target)
        assert isinstance(logp, Variable)

        # Value
        assert logp.data.numpy() == pytest.approx(expected_logp)

    def test_predict_mean_variance(self):
        """
        Test propagation of diagonal Gaussian through the likelihood density.
        """
        lik = self._standard_likelihood()
        input_mean = Variable(TensorType([0.0]))
        input_variance = Variable(TensorType([1.0]))
        expected_output_mean = input_mean
        expected_output_variance = input_variance + \
            self._expected_likelihood_variance
        
        # API
        output_mean, output_variance = lik.predict_mean_variance(input_mean,
            input_variance)
        assert isinstance(output_mean, Variable)
        assert isinstance(output_variance, Variable)

        # Value
        assert output_mean.data.numpy() == expected_output_mean.data.numpy()
        assert output_variance.data.numpy() == pytest.approx(
            expected_output_variance.data.numpy())

    def test_predict_mean_covariance(self):
        """
        Test propagation of full Gaussian through the likelihood density.
        """
        lik = self._standard_likelihood()
        input_mean = Variable(TensorType([0.0, 1.0, 2.1]))
        input_covariance = Variable(TensorType([[1.0, 0.5, 0.2], 
            [0.5, 1.0, 0.5], [0.2, 0.5, 1.0]]))
        expected_output_mean = input_mean
        # Ugh, sorry about this.  Will cleanup when we move PyTorch forward!
        expected_output_covariance = input_covariance + \
            Variable(TensorType([self._expected_likelihood_variance])).\
            expand_as(input_covariance).diag().diag()
        
        # API
        output_mean, output_covariance = lik.predict_mean_covariance(input_mean,
            input_covariance)
        assert isinstance(output_mean, Variable)
        assert isinstance(output_covariance, Variable)

        # Value
        assert all(output_mean.data.numpy() == expected_output_mean.data.numpy()
            )
        assert output_covariance.data.numpy() == pytest.approx(
            expected_output_covariance.data.numpy())

    @property
    def _expected_likelihood_variance(self):
        return 0.01

    def _standard_likelihood(self):
        return likelihoods.Gaussian(variance=self._expected_likelihood_variance)



#
# Apr 17, 2017  Yinhao Zhu
#
# class for probability density functions
from __future__ import absolute_import
from . import util as ut
from .functions import cholesky, inverse, cholesky_inverse

import torch as th
import numpy as np
from torch.autograd import Variable
from torch import DoubleTensor
import torch
from scipy.special import factorial
import scipy.stats


class Density(object):
    """
    Superclass for densities
    """
    # def __init__(self):
    #     ...

    def __init__(self, normalization=ut.as_variable(1.0)):
        self._dim = 0
        # _coefficient ensures that the pdf integrates to 1
        # Change normalization if you want an unnormalized density
        # (integrates to something other than 1)
        self._coefficient = 1.0
        self.normalization = normalization

    def pdf(self, x):
        raise NotImplementedError

    def log_pdf(self, x):
        # Override if needed (e.g. for better numerics)
        return torch.log(self.pdf(x))

    def sample(self, num_samples=1):
        raise NotImplementedError

    def get_dim(self):
        return self._dim

    def get_coefficient(self):
        return self._coefficient


class Exponential(Density):
    """
    Exponential distribution
    p(x; l) = l * exp(-x/l)

    mean = l
    variance = l^2
    """
    def __init__(self, length_scale):
        """
        Args:
            length_scale (torch.autograd.Variable): row vector of length scales
        """
        super(Exponential, self).__init__()
        if len(length_scale.size()) < 2:
            length_scale = length_scale.view(1, -1)
        self.length_scale = length_scale
        self._coefficient = 1.0 / th.prod(length_scale)
        self._dim = length_scale.size(1)

    def pdf(self, x):
        """
        Args:
            x (torch.autograd.Variable): argument
        Returns:
            (torch.autograd.Variable) the density at x
        """
        return self._coefficient * th.prod(th.exp(-x / self.length_scale))

    def log_pdf(self, x):
        return th.log(self._coefficient) + th.sum(-x / self.length_scale)

    def sample(self, num_samples=1):
        """
        Each row is a sample.

        :param num_samples:
        :return: (Variable)
        """
        return self.length_scale * Variable(DoubleTensor(
            np.random.exponential(size=(num_samples, self._dim))))


class InverseGamma(Density):
    """
    Inverse gamma distribution
    p(x; a, b) = b^(-1) / gamma(a) * x^(-a - 1) * exp(-1 / (b * x))
    a = shape
    b = rate (aka 1 / scale)

    Reference: https://en.wikipedia.org/wiki/Inverse-gamma_distribution

    mean = 1 / (b * (a - 1)) if a > 1
    mode = 1 / (b * (a + 1))
    variance = 1 / (b^2 * (a-1)^2 * (a-2)) if a > 2
    """

    def __init__(self, shape, rate, mean=None, variance=None):
        """
        Args:
            shape (torch.autograd.Variable): row vector of shape parameters
            rate (torch.autograd.Variable): row vector of rate parameters
        """
        super(InverseGamma, self).__init__()
        if mean is not None:
            assert variance is not None, "Must provide variance with mean"
            assert isinstance(mean, Variable), \
                "Provide mean as Pytorch Variable"
            assert isinstance(variance, Variable), \
                "Provide variance as Pytorch Variable"
            # Compute shape & rate from provided mean & var
            shape = th.pow(mean, 2) / variance + 2.0
            rate = th.pow(mean * (shape - 1.0), -1)
        assert isinstance(shape, Variable), "Provide shape as torch Variable"
        assert isinstance(rate, Variable), "Provide rate as torch Variable"
        if len(shape.size()) < 2:  # Given as 1d tensor
            shape = shape.view(1, -1)
        if len(rate.size()) < 2:  # Given as 1d tensor
            rate = rate.view(1, -1)
        self.shape = shape
        self.rate = rate
        self._coefficient = th.pow(rate, -shape) / \
                            Variable(th.exp(th.lgamma(shape.data)))
        self._dim = shape.size(1)

        # For RNG using gamma distribution:
        self._gamma_shape = self.shape.data.numpy().flatten()
        self._gamma_scale = self.rate.data.numpy().flatten()

    def pdf(self, x):
        """
        Note: not sure about scipy implementation, so coding up my own.

        Args:
            x (torch.autograd.Variable): argument
        Returns:
            (torch.autograd.Variable) the density at x
        """
        return th.prod(self._coefficient * th.pow(x, -self.shape - 1.0) * \
                       th.exp(-1.0 / (self.rate * x)))

    def sample(self, num_samples=1):
        """
        Each row is a sample.

        Strategy: use numpy's gamma-distributed rng, then transform

        :param num_samples:
        :return: (Variable)
        """
        return ut.as_variable(np.random.gamma(shape=self._gamma_shape,
                                              scale=self._gamma_scale,
                                              size=(num_samples, self._dim))
                              ** (-1))


class UniformCube(Density):
    def __init__(self, lower_bounds, upper_bounds):
        """
        Args:
            lower_bounds (torch.autograd.Variable): row vector of lower bounds
            upper_bounds (torch.autograd.Variable): row vector of upper bounds
        """
        super(UniformCube, self).__init__()
        if len(lower_bounds.size()) < 2:
            lower_bounds = lower_bounds.view(1, -1)
        if len(upper_bounds.size()) < 2:
            upper_bounds = upper_bounds.view(1, -1)

        assert lower_bounds.size(1) == upper_bounds.size(1), \
            "Sizes of lower and upper bounds do not match"

        self.lower_bounds = lower_bounds
        self.upper_bounds = upper_bounds
        self.range = upper_bounds - lower_bounds
        self._coefficient = 1.0 / th.prod(self.range)
        self._dim = lower_bounds.size(1)

    def pdf(self, x):
        """
        Args:
            x (torch.autograd.Variable): argument
        Returns:
            (torch.autograd.Variable, 1D) the density at x (
        """
        return self._coefficient * \
            torch.prod((x < self.upper_bounds) * (x >= self.lower_bounds), 1).\
            type(DoubleTensor)

    def sample(self, num_samples=1):
        """
        Each row is a sample.

        :param num_samples:
        :return: (Variable)
        """
        return self.lower_bounds + self.range * \
            ut.as_variable(np.random.rand(num_samples, self._dim))


# TODO
class GaussianUnivariate(Density):
    """
        Multivariate Gaussian distribution with diagonal covariance matrix
        p(x; m, s) = (2*pi)^(-d/2) |s|^(-1/2) *
                     exp(-1/2 sum_{i=1}^d ((x_i - m_i) / s_i)^2
        mean m and vector of standard deviations s

        mean = mu
        variance = sigma_var = self.sigma^2
        """

    def __init__(self, mu, sigma_var, normalization=ut.as_variable(1.0)):
        """
        Args:
            mu (torch.autograd.Variable): vector of means
            sigma_var (torch.autograd.Variable): vector of variances
        """
        super(GaussianMultivariateDiagonal, self).__init__(normalization)
        assert mu.numel() == sigma_var.numel(), \
            "Mean and variance vectors must be the same size"

        mu = mu.view(-1)
        sigma_std = torch.sqrt(sigma_var).view(-1)
        self.mu = mu
        self.sigma_std = sigma_std
        self._dim = mu.numel()

    def pdf(self, x):
        """
        Args:
            x (torch.autograd.Variable): argument.  Must be 2-dimensional
                matrix.
        Returns:
            (torch.autograd.Variable) the density at x
        """

        return self.normalization * ut.as_variable(np.array([np.prod(
            scipy.stats.norm.pdf(
                x_i.data.numpy(),
                self.mu.data.numpy(),
                self.sigma_std.data.numpy())
        ) for x_i in x]))

    def sample(self, num_samples=1):
        """
        Each row is a sample.

        :param num_samples:
        :return: (Variable)
        """
        return ut.as_variable(np.random.normal(
            size=(num_samples, self._dim),
            loc=self.mu.data.numpy().reshape((1, -1)),
            scale=self.sigma_std.data.numpy().reshape((1, -1))))


class GaussianMultivariateDiagonal(Density):
    """
    Multivariate Gaussian distribution with diagonal covariance matrix
    p(x; m, s) = (2*pi)^(-d/2) |s|^(-1/2) *
                 exp(-1/2 sum_{i=1}^d ((x_i - m_i) / s_i)^2
    mean m and vector of standard deviations s

    mean = mu
    variance = sigma_var = self.sigma^2
    """
    def __init__(self, mu, sigma_var, normalization=ut.as_variable(1.0)):
        """
        Args:
            mu (torch.autograd.Variable): vector of means
            sigma_var (torch.autograd.Variable): vector of variances
        """
        super(GaussianMultivariateDiagonal, self).__init__(normalization)
        assert mu.numel() == sigma_var.numel(), \
            "Mean and variance vectors must be the same size"

        mu = mu.view(-1)
        sigma_std = torch.sqrt(sigma_var).view(-1)
        self.mu = mu
        self.sigma_std = sigma_std
        self._dim = mu.numel()

    def pdf(self, x):
        """
        Args:
            x (torch.autograd.Variable): argument.  Must be 2-dimensional
                matrix.
        Returns:
            (torch.autograd.Variable) the density at x
        """

        return self.normalization * ut.as_variable(np.array([np.prod(
            scipy.stats.norm.pdf(
                x_i.data.numpy(),
                self.mu.data.numpy(),
                self.sigma_std.data.numpy())
            ) for x_i in x]))

    def sample(self, num_samples=1):
        """
        Each row is a sample.

        :param num_samples:
        :return: (Variable)
        """
        return ut.as_variable(np.random.normal(
            size=(num_samples, self._dim),
            loc=self.mu.data.numpy().reshape((1, -1)),
            scale=self.sigma_std.data.numpy().reshape((1, -1))))


class GaussianMultivariateFull(Density):
    """
    Multivariate Gaussian distribution with full covariance matrix
    p(x; m, S) = (2*pi)^(-d/2) |S|^(-1/2) *
                 exp(-1/2 (x-m)^T S^(-1) (x-m))
    mean m
    covariance matrix S

    mean = mu
    variance = sigma_var = self.sigma^2
    """
    def __init__(self, mu, sigma, normalization=ut.as_variable(1.0),
                 compute_precision=False, compute_chol=False):
        """
        Args:
            mu (torch.autograd.Variable): vector of means
            sigma_var (torch.autograd.Variable): vector of variances
        """
        super(GaussianMultivariateFull, self).__init__(normalization)
        assert mu.numel() == sigma.size(0), \
            "Mean and variance vectors must be the same size"

        self.mu = mu.view(-1)
        self.sigma = sigma
        self._l = None  # Upper Cholesky of sigma
        self._precision = None
        self._coefficient = None
        self._dim = mu.numel()
        if compute_precision:
            self._get_precision()
        if compute_chol:
            self._get_chol()

    def _get_precision(self):
        if self._l is None:
            self._get_chol()
        self._precision = Variable(cholesky_inverse(self._l.data.t()))

    def _get_chol(self):
        self._l = cholesky(self.sigma).t()

    def _get_coefficient(self):
        if self._l is None:
            self._get_chol()
        self._coefficient = (2.0 * np.pi) ** (-0.5 * self._dim) / \
                            torch.prod(self._l.diag())

    def pdf(self, x):
        """
        Args:
            x (torch.autograd.Variable): argument
        Returns:
            (torch.autograd.Variable) the density at x
        """
        if self._precision is None:
            self._get_precision()
        if self._coefficient is None:
           self._get_coefficient()

        return self.normalization * self._coefficient * torch.cat([
            torch.exp(-0.5 * torch.mm((xi - self.mu).view(1, -1),
                                      torch.mm(self._precision,
                                               (xi - self.mu).view(-1, 1))))
            for xi in x.split(1)]).view(-1)

    def sample(self, num_samples=1):
        """
        Each row is a sample.
        Stores the eigendecomposition of the covariance matrix so that
        subsequent sampling is very fast

        :param num_samples:
        :return: (Variable)
        """
        if self._l is None:
            self._get_chol()
        # mu + eval * randn(n, d) * evec
        return self.mu.view(1, -1) + torch.mm(
            Variable(torch.normal(torch.zeros(num_samples, self._dim),
                                  torch.ones(num_samples, self._dim)).
                type(DoubleTensor)),
            self._l)


def gaussianMD(x, mu, covar):
    """Calculates multi-dimension gaussian density
    Ref: PRML (Bishop) pg. 25

    TODO: make work for 1D also and delete gassian1D, also look into
    torch.normal

    Args:
        x (th.FloatTensor) = 1xD column tensor of the point of interest
        mu (th.FloatTensor) = 1xD column tensor of means 
        covar (th.FloatTensor) = DxD covariance matrix 
    """

    D = x.size()[0] #Dimension of gaussian
    if(D == 1):
        print('1D matrix detected, please use 1D gaussian method')
        return -1

    e = th.mm(th.transpose((x-mu),0,1), th.inverse(covar))
    e = th.mm(e, (x-mu))
    e = th.exp(-0.5*e)
    
    covar0 = Variable(covar).data.numpy()
    det = np.linalg.det(covar0)
    return 1.0/np.power(2*np.pi,D/2.0) * 1.0/(np.power(det,0.5)) * e


def gaussian1D(x, mu, var):
    """Calculates 1D gaussian density
    Ref: PRML (Bishop) pg. 24

    Args:
        x (Variable) = point of interest
        mu (Variable) = mean 
        var (Variable) = Variance squared 
    """
    e = (x-mu)*(1/var)
    e = e*(x-mu)
    e = np.exp(-0.5*e)

    return 1.0/(np.power(2*np.pi*var,0.5))*e


def gamma1D(x, a0, b0):
    """Calculates multi-dimension gamma density
    Ref: PRML (Bishop) pg. 100

    Args:
        x (Variable) = point of interest
        a0 (Variable) = gamma parameter 1 
        b0 (Variable) =  gamma parameter 2
    """
    n = 1/np.exp(ut.gammaln(a0))
    return  n * np.power(b0,a0) * np.power(x,a0-1) * np.exp(-b0 * x)


def studentT(x, mu, lam, nu):
    """Calculates multi-dimensional student T density
    Ref: PRML (Bishop) pg. 105

    **NOTE: it appears for large nu over 1e5 floating point precision starts to fail!**

    Args:
        x (th.FloatTensor) = 1xD column tensor of the point of interest
        mu (th.FloatTensor) = 1xD column tensor of means 
        lam (th.FloatTensor) = DxD precision matrix (inverse of covar)
        nu (Variable) = degrees of freedom
    """
    D = x.size()[0] #Dimension

    #Calculate constant out front
    c0 = np.exp(ut.gammaln((D + nu)/2.0) - ut.gammaln(nu/2.0))
    c0 = c0 * np.power(np.linalg.det(Variable(lam).data.numpy()),0.5) / np.power(np.pi*nu,D/2.0)
    #Mahalanobis Distance
    mah = th.mm(th.transpose((x-mu),0,1), lam)
    mah = th.mm(mah,(x-mu))
    #Tensor calculations
    t0 = th.add(th.div(mah,nu),1)
    t0 = th.pow(t0,(-(D+nu)/2.0))

    return th.mul(t0,c0)


def gaussian(x, mu, var):
    return float(-0.5 * np.log(2 * np.pi)) - 0.5 * th.log(var) \
           - 0.5 * th.pow(x - mu, 2) / var




#
# Yinhao Zhu, May 01, 2017
#
"""
Sparse GP regression, including variational GP and others.
"""

from __future__ import absolute_import

import torch
import numpy as np
from torch.utils.data import TensorDataset, DataLoader

from ..model import GPModel, Param
from ..functions import cholesky, trtrs
from ..mean_functions import Zero
from ..likelihoods import Gaussian
from ..util import TensorType, torch_dtype, as_tensor, kmeans_centers
from ..util import KL_Gaussian


class _InducingPointsGP(GPModel):
    """
    Parent class for GPs with inducing points
    """
    def __init__(self, y, x, kernel, num_inducing_points=None, 
            inducing_points=None, mean_function=None, likelihood=Gaussian()):
        """
        Assume Gaussian likelihood

        Args:
            observations (np.ndarray): Y, n x p
            input (np.ndarray): X, n x q
            kernel (gptorch.Kernel):
            inducing_points (np.ndarray, optional): Z, m x q
            num_inducing (int), optional): number of inducing inputs

        Input, observations, and kernel must be specified, if both
        ``inducing_points`` and ``num_inducing`` are not set, 1/10 th of total
        points (up to 100) will be draw randomly from input as the inducing 
        points.
        """
        
        super().__init__(y, x, kernel, likelihood, mean_function)

        if inducing_points is None:
            if num_inducing_points is None:
                num_inducing_points = np.clip(x.shape[0] // 10, 1, 100)
            inducing_points = kmeans_centers(x, num_inducing_points)
            # indices = np.random.permutation(len(x))[:num_inducing_points]
            # inducing_points = TensorType(x[indices])
            print("Inducing points:\n{}".format(inducing_points))
        
        # Z stands for inducing input points as standard in the literature
        self.Z = Param(as_tensor(inducing_points))
        self.jitter = 1.0e-6
        

class FITC(_InducingPointsGP):
    """
    Fully Independent Training Conditional approximation for GP

    References:
        Snelson, Edward, and Zoubin Ghahramani. "Sparse Gaussian processes
         using pseudo-inputs." Advances in neural information processing
         systems 18 (2006): 1257.
        Quinonero-Candela, Joaquin, and Carl Edward Rasmussen. "A unifying
        view of sparse approximate Gaussian process regression." Journal of
        Machine Learning Research 6.Dec (2005): 1939-1959.
    """
    # TODO: add FITC for sparse GP regression
    pass


class VFE(_InducingPointsGP):
    """
    Variational Free Energy approximation for GP

    Reference:
        Titsias, Michalis K. "Variational Learning of Inducing Variables
        in Sparse Gaussian Processes." AISTATS. Vol. 5. 2009.
    """
    def compute_loss(self):
        """
        Computes the variational lower bound of the true log marginal likelihood
        Eqn (9) in Titsias, Michalis K. "Variational Learning of Inducing Variables
        in Sparse Gaussian Processes." AISTATS. Vol. 5. 2009.
        """

        num_inducing = self.Z.size(0)
        num_training = self.X.size(0)
        dim_output = self.Y.size(1)
        # TODO: add mean_functions
        # err = self.Y - self.mean_function(self.X)
        err = self.Y
        Kff_diag = self.kernel.Kdiag(self.X)
        Kuf = self.kernel.K(self.Z, self.X)
        # add jitter
        Kuu = self.kernel.K(self.Z) + self.jitter * torch.eye(num_inducing, 
            dtype=torch_dtype)
        L = cholesky(Kuu)

        A = trtrs(Kuf, L)
        AAT = A @ A.t() / self.likelihood.variance.transform().expand_as(Kuu)
        B = AAT + torch.eye(num_inducing, dtype=torch_dtype)
        LB = cholesky(B)
        # divide variance at the end
        c = trtrs(A @ err, LB) / self.likelihood.variance.transform()

        # Evidence lower bound
        elbo = TensorType([-0.5 * dim_output * num_training * np.log(2*np.pi)])
        elbo -= dim_output * LB.diag().log().sum()
        elbo -= 0.5 * dim_output * num_training * self.likelihood.variance.transform().log()
        elbo -= 0.5 * (err.pow(2).sum() + dim_output * Kff_diag.sum()) \
                / self.likelihood.variance.transform()
        elbo += 0.5 * c.pow(2).sum()
        elbo += 0.5 * dim_output * AAT.diag().sum()

        return - elbo

    def _predict(self, input_new: TensorType, diag=True):
        """
        Compute posterior p(f*|y), integrating out induced outputs' posterior.

        :return: (mean, var/cov)
        """

        z = self.Z
        z.requires_grad_(False)

        num_inducing = z.size(0)
        dim_output = self.Y.size(1)

        # err = self.Y - self.mean_function(self.X)
        err = self.Y
        Kuf = self.kernel.K(z, self.X)
        # add jitter
        Kuu = self.kernel.K(z) + self.jitter * torch.eye(num_inducing, 
            dtype=torch_dtype)
        Kus = self.kernel.K(z, input_new)
        L = torch.cholesky(Kuu)
        A = trtrs(Kuf, L)
        AAT = A @ A.t() / self.likelihood.variance.transform().expand_as(Kuu)
        B = AAT + torch.eye(num_inducing, dtype=torch_dtype)
        LB = torch.cholesky(B)
        # divide variance at the end
        c = trtrs(A @ err, LB) / self.likelihood.variance.transform()
        tmp1 = trtrs(Kus, L)
        tmp2 = trtrs(tmp1, LB)
        mean = tmp2.t() @ c

        if diag:
            var = self.kernel.Kdiag(input_new) - tmp1.pow(2).sum(0).squeeze() \
                  + tmp2.pow(2).sum(0).squeeze()
            # add kronecker product later for multi-output case
        else:
            var = self.kernel.K(input_new) + tmp2.t() @ tmp2 - tmp1.t() @ tmp1
        # return mean + self.mean_function(input_new), var
        return mean, var


class SVGP(_InducingPointsGP):
    pass
    # """
    # Sparse variational Gaussian process.

    # Sparse GP with 

    # James Hensman, Nicolo Fusi, and Neil D. Lawrence,
    # "Gaussian processes for Big Data" (2013)

    # James Hensman, Alexander Matthews, and Zoubin Ghahramani, 
    # "Scalable variational Gaussian process classification", JMLR (2015).
    # """
    # def __init__(self, x, y, kernel, num_inducing_points=None, 
    #         inducing_points=None, mean_function=None, likelihood=Gaussian(), 
    #         batch_size=None):
    #     """
    #     :param batch_size: How many points to process in a minibatch of 
    #         training.  If None, no minibatches are used.
    #     """
    #     super().__init__(x, y, kernel, num_inducing_points=num_inducing_points, 
    #         inducing_points=inducing_points, mean_function=mean_function, 
    #         likelihood=likelihood)
    #     assert batch_size is None, "Minibatching not supported yet."
    #     self.batch_size = batch_size

    #     # Parameters for the Gaussian variational posterior over the induced
    #     # outputs:
    #     self.induced_output_mean, self.induced_output_chol_cov = \
    #         self._init_posterior()

    # def compute_loss(self, x: TensorType=None, y: TensorType=None) \
    #         -> TensorType:
    #     """
    #     :param x: batch inputs
    #     :param y: batch outputs
    #     """
    #     x, y = self._get_batch(x, y)
    #     qu_mean = self.induced_output_mean
    #     qu_lc = self.induced_output_chol_cov.transform()
    #     m = self.Z.shape[0]

    #     # Get the mean of the marginal q(f)
    #     k_uf = self.kernel.K(self.Z, x)
    #     kuu = self.kernel.K(self.Z) + \
    #         self.jitter * torch.eye(m, dtype=torch_dtype)
    #     kuu_chol = cholesky(kuu)
    #     a = torch.trtrs(k_uf, kuu_chol, upper=False)[0]
    #     f_mean = a.t() @ \
    #         torch.trtrs(qu_mean, kuu_chol, upper=False)[0]

    #     # Variance of the marginal q(f)
    #     b = torch.trtrs(q_cov_chol, kuu_chol, upper=False)[0]
    #     f_var_1 = (a.t() @ a).sum(1)
    #     f_var_2 = (a.t() @ b).sum(1)
    #     f_var = (self.kernel.Kdiag(x) + f_var_1 + f_var_2)[:, None].expand(
    #         *y.shape)

    #     elbo = self.likelihood.propagate(torch.distributions.Normal(f_mean, 
    #         f_var.sqrt()))
        
    #     kl = KL_Gaussian(qu_mean, qu_lc @ qu_lc.t(), 
    #         torch.zeros(*qu_mean.shape, dtype=torch_dtype, kuu)
    #     return elbo - kl
        
    # def _predict(self, input_new):
    #     raise NotImplementedError("")
    
    # def _get_batch(self, x, y):
    #     """
    #     Get the next batch of data for training.
    #     :return: (TensorType, TensorType) inputs, outputs
    #     """
    #     assert not ((x is None) ^ (y is None)), \
    #         "Cannot provide inputs or outputs only in minibatch"
    #     return self.X, self.Y if x is None else x, y

    # def _init_posterior(self):
    #     """
    #     Get an initial guess at the variational posterior over the induced 
    #     outputs.
    #     """
    #     # For the mean, take the nearest points in input space and steal their
    #     # corresponding outputs.  This could be costly if X is very large...
    #     nearest_points = np.argmin(
    #         squared_distance(self.Z, self.X).detach().numpy(), axis=1)
    #     mean = self.X[nearest_points]

    #     # For the covariance, we'll start with 1/100 of the prior kernel
    #     # matrix. (aka 1/10th the Cholesky).
    #     cov = 0.1 * cholesky(self.kernel.K(self.Z))
    #     return mean, cov



# File: test_gpr.py
# File Created: Sunday, 21st July 2019 7:53:26 pm
# Author: Steven Atkinson (steven@atkinson.mn)

"""
Tests for the GPR class
"""

import pytest


def TestGPR(object):
    # TODO
    pass



from __future__ import absolute_import
from .util import as_variable
from .model import Model, Param
from torch.autograd import Variable
import torch as th
from numpy.polynomial.hermite import hermval
import numpy as np


TensorType = th.DoubleTensor


class MeanFunction(Model):
    """
    The base mean function class.
    To implement a mean function, write the __call__ method. This takes a
    tensor X and returns a tensor m(X). In accordance with the GPflow/GPtorch
    standard, each row of X represents one datum, and each row of Y is computed
    independently for each row of X.

    MeanFunction classes can have parameters, see the Linear class for an
    example.
    """
    def __init__(self):
        super(MeanFunction, self).__init__()

    def __call__(self, X):
        raise NotImplementedError("Implement the __call__\
                                  method for this mean function")

    def __add__(self, other):
        return Additive(self, other)

    def __mul__(self, other):
        return Product(self, other)

    def __repr__(self):
        tmpstr = self.__class__.__name__ + ' (\n\n'
        for name, param in self._parameters.items():
            tmpstr = tmpstr + name + str(param.data) + '\n'
        tmpstr = tmpstr + ')'
        return tmpstr


class Zero(MeanFunction):
    def __call__(self, X):
        # return Variable(th.zeros(len(X), 1), requires_grad=False)
        # print(th.zeros(10, 1))
        return Variable(th.zeros(X.size(0), 1).type(TensorType),
                        requires_grad=False)


class Constant(MeanFunction):
    """
    Just a constant
    """
    def __init__(self):
        super(Constant, self).__init__()

    def __call__(self, x):
        if isinstance(x, Variable):
            return Variable(th.ones(x.size(0), 1).type(TensorType))
        else:
            raise NotImplementedError


class AskeyPolynomial(MeanFunction):
    """
    Superclass for computing the tensor product of polynomials from the Askey
    scheme
    """
    def __init__(self, degree, tensor_product=False):
        """

        Args:
            degree (int): degree of polynomials
            tensor_product (bool): Whether to take the tensor product of the 1D
                basis functions.
                WARNING: for Grid inputs, the Kronecker structure of the design
                    matrix means that you HAVE to take the tensor product of the
                    basis functions associated with the various subgrids.
                    Unforutnately, this prevent an apples-to-apples comparison
                    with a normal full input (unless we define specific tensor
                    products to be taken.  Perhaps we can do this in the
                    future...)
        """
        super(AskeyPolynomial, self).__init__()
        self.f = None  # Replace with the appropriate function in subclasses
        self.degree = Param(th.IntTensor([degree]), requires_transform=False,
                            requires_grad=False)
        self.tensor_product = Param(th.ByteTensor([tensor_product]),
                                    requires_transform=False,
                                    requires_grad=False)

    def __call__(self, x):
        return self._eval(x)

    def _eval(self, x):
        """
        Evaluates the tensor product of the Hermite polynomials evaluated on X's
        columns.

        Args:
            x (Variable): columns are arguments to the Hermite polynomials
        Returns:
            (Variable)
        """
        assert(isinstance(x, Variable)), "Need Variable input"
        assert(len(x.size()) == 2), "Need a matrix as input"
        if self.tensor_product.data[0]:
            return th.cat([KroneckerProduct([
                self._eval_scalar(x[i, j]).view(1, -1)
                for j in range(x.size(1))]).eval()
                for i in range(x.size(0))])
        else:
            return th.cat([th.cat([
                self._eval_scalar(x[i, j], j == 0).view(1, -1)
                for j in range(x.size(1))], 1)
                for i in range(x.size(0))], 0)

    def _eval_scalar(self, x, include_const=True):
        """
        Evaluate polynomial terms on 1D inputs
        Args:
            x (Variable) a scalar
            include_const (bool): include the constant term or not.
        Returns:
            (Variable) A 1D tensor (length = degree + 1) of its Hermite
                polynomials.
        """
        if include_const:
            range_start = 0
        else:
            range_start = 1
        y_array = np.array([
            self.f(x.data[0], np.concatenate((np.zeros(i), [1])))
            for i in range(range_start, self.degree.data[0] + 1)])
        y = as_variable(y_array)
        return y


class HermitePolynomial(AskeyPolynomial):
    """
    Computes the tensor product of the Hermite polynomials on the columns of a
    matrix.

    This is the natural basis to use in gPC when the stochastic variables are
    Gaussian-distributed on (-inf, inf).

    Hermite polynomials:
    https://en.wikipedia.org/wiki/Hermite_polynomials#Definition

    Note: following numpy's implementation, these are the "Physicists'
    polynomials" according to Wikipedia
    """
    def __init__(self, degree, tensor_product=True):
        super(HermitePolynomial, self).__init__(degree, tensor_product)
        self.f = np.polynomial.hermite.hermval


class LegendrePolynomial(AskeyPolynomial):
    """
    Computes the tensor product of the Legendre polynomials on the columns of a
    matrix.

    This is the natural basis to use in gPC when the stochastic variables are
    uniformly-distributed on some finite interval.

    Legendre polynomials:
    https://en.wikipedia.org/wiki/Legendre_polynomials
    """
    def __init__(self, degree, tensor_product=False):
        super(LegendrePolynomial, self).__init__(degree, tensor_product)
        self.f = np.polynomial.legendre.legval


class LaguerrePolynomial(AskeyPolynomial):
    """
    Computes the tensor product of the Legendre polynomials on the columns of a
    matrix.

    This is the natural basis to use in gPC when the stochastic variables are
    exponentially-distributed on [0, inf)

    Reference:
    https://en.wikipedia.org/wiki/Laguerre_polynomials#The_first_few_polynomials
    """
    def __init__(self, degree, tensor_product=True):
        super(LegendrePolynomial, self).__init__(degree, tensor_product)
        self.f = np.polynomial.laguerre.lagval


class Additive(MeanFunction):
    def __init__(self, first_part, second_part):
        MeanFunction.__init__(self)
        self.add_1 = first_part
        self.add_2 = second_part

    def __call__(self, X):
        return th.add(self.add_1(X), self.add_2(X))


class Product(MeanFunction):
    def __init__(self, first_part, second_part):
        MeanFunction.__init__(self)

        self.prod_1 = first_part
        self.prod_2 = second_part

    def __call__(self, X):
        return th.mul(self.prod_1(X), self.prod_2(X))



# File: test_kernels.py
# File Created: Sunday, 11th November 2018 1:31:17 pm
# Author: Steven Atkinson (steven@atkinson.mn)
#
# Tests:
# K(x)
# K(x, x2)
# Transpose
# Shift (Stationary only)

from unittest import TestCase
import pytest
import numpy as np
from torch.autograd import Variable
import os

from gptorch import kernels
from gptorch.util import TensorType

data_dir = os.path.join(os.path.dirname(__file__), "data", "kernels")


def as_variable(x: np.array) -> Variable:
    return Variable(TensorType(x))


class Kern(object):
    def setUp(self, kernel_type):
        self.kernel_type = kernel_type
        self.x1 = as_variable(np.load(os.path.join(data_dir, "x1.npy")))
        self.x2 = as_variable(np.load(os.path.join(data_dir, "x2.npy")))
        self.n1, self.d1 = self.x1.data.numpy().shape
        self.n2, self.d2 = self.x2.data.numpy().shape
        self.kern = self.kernel_type(self.d1)
        self.kern_str = self.kern.__class__.__name__
        self.kx_expected = np.load(os.path.join(data_dir, "{}_kx.npy".format(
            self.kern_str)))
        self.kx2_expected = np.load(os.path.join(data_dir, "{}_kx2.npy".format(
            self.kern_str)))
        self.kdiag_expected = np.load(os.path.join(data_dir, 
            "{}_kdiag.npy".format(self.kern_str)))

    def test_add(self):
        """
        Test kernel addition operator
        """
        k1 = self.kern + self.kern
        k2 = kernels.Sum(self.kern, self.kern)
        k1x = k1.K(self.x1)
        k2x = k2.K(self.x1)
        assert all([x1 == x2 for x1, x2 in zip(k1x.flatten(), k2x.flatten())])

    def test_mul(self):
        """
        Test kernel product operator
        """
        k1 = self.kern * self.kern
        k2 = kernels.Product(self.kern, self.kern)
        k1x = k1.K(self.x1)
        k2x = k2.K(self.x1)
        assert all([x1 == x2 for x1, x2 in zip(k1x.flatten(), k2x.flatten())])

    def test_K(self):
        """
        Inputs are expected to be torch.autograd.Variables of 
        torch.DoubleTensors

        Output is expected to be a torch.autograd.Variable
        """
        kx_actual = self.kern.K(self.x1).data.numpy()
        kx2_actual = self.kern.K(self.x1, self.x2).data.numpy()
        kx2t_actual = self.kern.K(self.x2, self.x1).data.numpy()
        assert np.allclose(self.kx_expected, kx_actual)
        assert np.allclose(self.kx2_expected, kx2_actual)

        # Test symmetric K()
        assert np.allclose(kx_actual.T, kx_actual)

        # Test transpose of cross-kernel:
        assert np.allclose(self.kx2_expected, kx2t_actual.T)

    def test_Kdiag(self):
        kdiag_actual = self.kern.Kdiag(self.x1).data.numpy()
        assert np.allclose(self.kdiag_expected, kdiag_actual)


class Stationary(Kern):
    def setUp(self, kernel_type):
        super().setUp(kernel_type)
        x_shift = 0.34
        self.x1_shift = self.x1 + x_shift
        self.x2_shift = self.x1 + x_shift

    def test_K(self):
        super().test_K()
        kx_shift_actual = self.kern.K(self.x1_shift).data.numpy()
        assert np.allclose(self.kx_expected, kx_shift_actual)

    def test_Kdiag(self):
        super().test_Kdiag()
        kxdiag_shift_actual = self.kern.Kdiag(self.x1_shift).data.numpy()
        assert np.allclose(self.kdiag_expected, kxdiag_shift_actual)


class ARD(Stationary):
    def setUp(self, kernel_type):
        super().setUp(kernel_type)
        self.ard_length_scales = np.load(os.path.join(data_dir, 
            "ard_length_scales.npy"))
        self.kern_ard = self.kernel_type(self.d1, ARD=True, 
            length_scales=self.ard_length_scales)
        self.kx_ard_expected = np.load(os.path.join(data_dir, 
            "{}_kx_ard.npy".format(self.kern_str)))
        self.kx2_ard_expected = np.load(os.path.join(data_dir, 
            "{}_kx2_ard.npy".format(self.kern_str)))
        self.kdiag_ard_expected = np.load(os.path.join(data_dir,
            "{}_kdiag_ard.npy".format(self.kern_str)))

    def test_K(self):
        super().test_K()
        kx_ard_actual = self.kern_ard.K(self.x1).data.numpy()
        kx2_ard_actual = self.kern_ard.K(self.x1, self.x2).data.numpy()
        assert np.allclose(self.kx_ard_expected, kx_ard_actual)
        assert np.allclose(self.kx2_ard_expected, kx2_ard_actual)

    def test_Kdiag(self):
        super().test_Kdiag()
        kdiag_ard_actual = self.kern_ard.Kdiag(self.x1).data.numpy()
        assert np.allclose(self.kdiag_ard_expected, kdiag_ard_actual)


class TestWhite(Kern, TestCase):
    def setUp(self):
        super().setUp(kernels.White)    


class TestConstant(Kern, TestCase):
    def setUp(self):
        super().setUp(kernels.Constant) 


class TestBias(Kern, TestCase):
    def setUp(self):
        super().setUp(kernels.Bias)


class TestExp(ARD, TestCase):
    def setUp(self):
        super().setUp(kernels.Exp)


class TestMatern12(ARD, TestCase):
    def setUp(self):
        super().setUp(kernels.Matern12)


class TestMatern32(ARD, TestCase):
    def setUp(self):
        super().setUp(kernels.Matern32)


class TestMatern52(ARD, TestCase):
    def setUp(self):
        super().setUp(kernels.Matern52)


class TestRbf(ARD, TestCase):
    def setUp(self):
        super().setUp(kernels.Rbf)


@pytest.mark.xfail(reason="Linear is different because we use .variance " + 
    "instead of .length_scales (TODO)")
class TestLinear(ARD, TestCase):
    def setUp(self):
        super().setUp(kernels.Linear)



#
# May 10, 2017  Yinhao Zhu
# Utilities: data loading, plotting, batching...
from torch.autograd import Variable
import torch
import numpy as np
from warnings import warn
from scipy.cluster.vq import kmeans2


TensorType = torch.DoubleTensor
torch_dtype = torch.double


gamma_cof=[57.1562356658629235,-59.5979603554754912,
    14.1360979747417471,-0.491913816097620199,.339946499848118887e-4,
    .465236289270485756e-4,-.983744753048795646e-4,.158088703224912494e-3,
    -.210264441724104883e-3,.217439618115212643e-3,-.164318106536763890e-3,
    .844182239838527433e-4,-.261908384015814087e-4,.368991826595316234e-5]


def tensor_type():
    warn("Use TensorType instead of tensor_type()", DeprecationWarning)
    return TensorType


def as_variable(x):
    """
    Convert a numpy matrix to a Tensor variable

    Args:
        x (np.ndarray)
    Returns:
        (TensorType)
    """
    warn("Deprecated.  Use as_tensor() instead.", DeprecationWarning)
    return as_tensor(x)


def as_tensor(x):
    """
    Convert a numpy matrix into a raw Tensor.
    If you're unsure, use as_variable() instead.

    Args:
        x (np.ndarray)
    Returns:
        (TensorType)
    """
    if isinstance(x, torch.Tensor):
        return x.type(TensorType)
    if isinstance(x, np.ndarray):
        return TensorType(x)
    elif isinstance(x, float):
        return TensorType([x])
    else:
        raise TypeError("Unsupported type {}".format(type(x)))


def KL_Gaussian(m1, m2, S1, S2):
    """
    KL-divergence between two Gaussians
    $p\sim\mathcal{N}(m1, S1), q\sim\mathcal{N}(m2, S2)$
    $\textrm{KL}(p||q) = \frac{1}{2}(\log\frac{|\Sigma_2|}{\Sigma_1} - n + \text{Tr}((\Sigma_2^{-1}\Sigma_1) + (m_2-m_1)^T\Sigma_2^{-1}(m_2-m_1)$
    :param m1:
    :param m2:
    :param S1:
    :param S2:
    :return:
    """
    raise NotImplementedError("")


def kmeans_centers(x: np.ndarray, k: int) -> np.ndarray:
    """
    Use k-means clustering and find the centers of the clusters.
    :param x: The data
    :param k: Number of clusters
    :return: the centers
    """
    return kmeans2(x, k)[0]


def PCA(X, q):
    """
    A helpful function for linearly reducing the dimensionality of the data X
    to q.

    Args:
        X (np.ndarray(n, p)): Observations
        q (int): Dimensionality of latent variable

    Returns:
        (np.ndarray(n, q)): PCA projection array of Z

    """
    assert q <= X.shape[1], 'Cannot have more latent dimensions than observed'
    evals, evecs = np.linalg.eigh(np.cov(X.T))
    indices = np.argsort(evals)[::-1]
    W = evecs[:, indices]
    W = W[:, :q]
    return (X - X.mean(0)).dot(W)


def gammaln(xx):
    """
    Used to approximate the gamma function using the Lanczos algorithm
    Ref: Numerical Methods (Third Edition) by William Press et al. pg. 256

    Args:
        xx (Variable): desired gamma value, must be > 0
    Returns:
        ln(gamma(xx)) (Variable): natural log of desired gamma value
    """

    if (xx <= 0):
        raise ValueError('Illegal value for gammaln!')
    y = x = xx
    tmp = x + 5.24218750000000000    #Rational 671/128.
    tmp = (x+0.5)*np.log(tmp)-tmp
    ser = 0.999999999999997092 #First coefficiect
    for i in range(0,14):
        ser += gamma_cof[i]/(y + i + 1)
    return tmp+np.log(2.5066282746310005*ser/x)


def squared_distance(x1: TensorType, x2: TensorType=None) -> TensorType:
    """
    Given points x1 [n1 x d1] and x2 [n2 x d2], return a [n1 x n2] matrix with
    the pairwise squared distances between the points.
    """
    if x2 is None:
        return squared_distance(x1, x1)
    x1s = x1.pow(2).sum(1, keepdim=True)
    x2s = x2.pow(2).sum(1, keepdim=True)
    # Prevent negative squared distances
    return torch.clamp(x1s + x2s.t() -2.0 * x1 @ x2.t(), min=0.0)



# Yinhao Zhu
# yzhu10@nd.edu
# April 16, 2017  

"""
Basic model and parameter classes for Gaussian Processes, inheriting from
:class:`torch.nn.Module` and :class:`torch.nn.Parameter` respectively.
"""

from __future__ import absolute_import

from torch.autograd import gradcheck
import torch
from scipy.optimize import minimize
import numpy as np
from time import time
from warnings import warn

from .functions import SoftplusInv, cholesky
from .util import TensorType


def input_as_tensor(predict_func):
    """
    Decorator for prediction funtions to ensure that inputs are TensorType
    objects before passing into GPModel._predict() methods.

    :param predict_func: The public predict funciton to be wrapped
    """
    def predict(obj, input_new, *args, **kwargs):
        if isinstance(input_new, np.ndarray):
            input_new = TensorType(input_new)
        return predict_func(obj, input_new, *args, **kwargs)
    
    return predict


def _addindent(s_, numSpaces):
    s = s_.split('\n')
    # dont do anything for single-line stuff
    if len(s) == 1:
        return s_
    first = s.pop(0)
    s = [(numSpaces * ' ') + line for line in s]
    s = '\n'.join(s)
    s = first + '\n' + s
    return s


class Model(torch.nn.Module):
    """
    Customized Model class for all GP objects
    """
    def forward(self):
        return None

    def __repr__(self):
        tmpstr = self.__class__.__name__ + ' (\n'
        for name, param in self._parameters.items():
            tmpstr = tmpstr + name + '\n' + \
                     str(param.transform().data)  + '\n'
        for key, module in self._modules.items():
            modstr = module.__repr__()
            modstr = _addindent(modstr, 2)
            tmpstr = tmpstr + '  (' + key + '): ' + modstr + '\n'
        tmpstr = tmpstr + ')' + '\n'
        return tmpstr

    # Three functions wrap the scipy.optimize.minimize
    # get_param_array
    # set_parameters
    # loss_and_grad
    def _get_param_array(self):
        """Returns a 1D array by flattening and concatenating all the parameters
        in the model.
        """
        param_array = []
        for param in self.parameters():
            if param.requires_grad:
                param_array.append(param.data.numpy().flatten())

        return np.concatenate(param_array)

    def _set_parameters(self, param_array):
        """Set the parameters from a parameter array in the format of
        the return of get_param_array().
        """
        # assert isinstance(param_array, np.ndarray)
        idx_current = 0
        for param in self.parameters():
            if param.requires_grad:
                idx_next = idx_current + np.prod(param.data.size())
                param_np = np.reshape(param_array[idx_current: idx_next], param.data.numpy().shape)
                idx_current = idx_next
                param.data = TensorType(param_np)

    def _loss_and_grad(self, param_array):
        """
        Computes the loss and gradients for the parameters with values
        ```param_array```.

        1) Take param_array and expand it into the model
        2) Compute the loss and grad
        3) Put the results into numpy data structures

        Note: Any grad entries that are infinite are replaced with zeroes.

        Args:
            param_array (np.ndarray): the state x

        Returns:
            f(x) (np.float64), the objective function values
            g(x) (np.ndarray of np.float64's), the gradient
        """

        # 1) set parameters of the model from a 1D param_array
        self._set_parameters(param_array)
        # print(self)
        for name, param in self.named_parameters():
            # param.grad will be None if either (1) this is the first time we've
            # visited or (2) this param doesn't require a gradient
            if param.grad is not None:
                param.grad.data.zero_()

        # 2) Compute
        loss = self.compute_loss()
        loss.backward()

        # 3) Return as the proper numpy types
        grad = []
        for name, param in self.named_parameters():
            if param.requires_grad:
                grad.append(param.grad.data.numpy().flatten())
            # if name in ['Z']:
            #     print('Z: %s' % param.data)
            #     print('grad of Z: %s' % param.grad)
        print('loss: %s' % loss.data.numpy())
        grad = np.concatenate(grad)
        grad_isfinite = np.isfinite(grad)
        ## scipy.optimizer.minimize.L-BFGS-B expects double precision (Fortran)
        if np.all(grad_isfinite):
            return loss.data.numpy().astype(np.float64), grad.astype(np.float64)
        else:
            # self._previous_x = x  # store the last known good value
            print("Warning: inf or nan in gradient: replacing with zeros")
            return loss.data.numpy().astype(np.float64), \
                   np.where(grad_isfinite, grad, 0.).astype(np.float64)

    # Functions for using gradcheck on your model.
    # Use self.loss as the function to be grad-checked, and provide the inputs
    # with self.extract_params
    def extract_params(self):
        """
        Returns:
            (Tuple of TensorTypes)
        """
        return tuple([x for x in self.parameters()])

    def expand_params(self, *args):
        """
        Args:
            args (tuple of TensorTypes): (Get from self.extract_params())
        """
        for arg, (param_name, param) in zip(args, self.named_parameters()):
            if isinstance(arg, Param):
                param.data = arg.data
            elif isinstance(arg, np.ndarray):
                raise NotImplementedError(
                    "Unresolved issues with expanding numpy arrays")

    def loss(self, *args):
        """
        Loss, given args to be expanded into the model.

        Args:
            args (pointer to a tuple of TensorTypes):
            (Get from self.extract_params())
        """
        self.expand_params(*args)
        return self.compute_loss()

    def gradcheck(self, eps=1e-6, atol=1e-5, rtol=1e-3, verbose=False):
        """
        This performs a gradcheck on your model's self.compute_loss() function

        See torch.autograd.gradcheck for more on eps, atol, and rtol.

        Args:
            eps (float): Finite difference step size
            atol (float): absolute tolerance (takes over with grads near zero)
            rtol (float): relative tolerance (takes over with large grads)
            verbose (bool): Output more
        Returns:
             (bool) whether the gradcheck passed (true = good)
        """
        if verbose:
            warn("Verbose not yet figured out")
        return gradcheck(self.loss, self.extract_params(), eps=eps, atol=atol,
                         rtol=rtol)


class Param(torch.nn.Parameter):
    """
    Customized Parameter class extending the PyTorch Parameter class.
    Its main purpose is to include the following additional functionality:
    1) The .transform() member function, in order to impose constraints on the 
        parameter.
    2) the .prior member, for incorporation into joint log-probabilities (e.g. 
        for training)
    """
    def __new__(cls, data=None, requires_grad=True, requires_transform=False,
                prior=None):
        if requires_transform:
            data = Param._transform_log(data, forward=False)
        return super(Param, cls).__new__(cls, data, 
            requires_grad=requires_grad)

    def __init__(self, data, requires_grad=True, requires_transform=False):
        super(Param, self).__init__()
        self.requires_transform = requires_transform
        self.prior = None

    def transform(self):
        # Avoid in-place operation for Tensor, using clone method  ???
        if self.requires_transform:
            return self._transform_log(self.clone(), forward=True)
        else:
            return self

    def __repr__(self):
        return 'Parameter containing:' + self.data.__repr__()

    @staticmethod
    def _transform_log(x, forward):
        if forward:
            return torch.exp(x)
        else:
            return torch.log(x)

    @staticmethod
    def _transform_softplus(x, forward):
        if forward:
            return torch.nn.functional.softplus(x, threshold=35)
        else:
            return SoftplusInv(x)


class GPModel(Model):
    """
    The base class for GP models
    """

    def __init__(self, y, x, kernel, likelihood, mean_function, 
                 name='gp'):
        """
        For unsupervised case, x is optional ...

        Args:
            y (ndarray): Y, N x q
            x (ndarray): X, N x p
            kernel (gptorch.Kernel):
            likelihood (gptorch.Likelihood):
            mean_function (gptorch.MeanFunction):
            name (string): name of this model
        """
        super().__init__()
        self.kernel = kernel
        self.likelihood = likelihood
        # TODO: reworking mean functions as parameterized classes instead of
        # design & weight matrices.
        assert mean_function is None, "Mean functions not supported"
        self.mean_function = mean_function

        allowed_data_types = (np.ndarray, TensorType)
        assert type(x) in allowed_data_types, \
            "x must be one of {}".format(allowed_data_types)
        if isinstance(x, np.ndarray):
            x = TensorType(x)
        assert type(y) in allowed_data_types, \
            "y must be one of {}".format(allowed_data_types)
        if isinstance(y, np.ndarray):
            y = TensorType(y)
        x.requires_grad_(False)
        y.requires_grad_(False)
        self.X, self.Y = x, y
        self.__class__.__name__ = name

    def compute_loss(self):
        """
        Defines computation graph upon every call - PyTorch

        This function must be implemented by all subclasses.
        """
        raise NotImplementedError

    # def forward(self, *data):
    #     # Builds computation graph for an objective, e.g. log marginal likelihood
    #     # Responds to the NotImplementedError in super class: torch.nn.Module
    #     # data usually includes observations and input
    #     return self.compute_likelihood(*data)

    def optimize(self, method='Adam', max_iter=2000, verbose=True,
                 learning_rate=None):

        """
        Optimizes the model by minimizing the loss (from :method:) w.r.t.
        model parameters.

        Args:
            method (torch.optim.Optimizer, optional): Optimizer in PyTorch
                (maybe add scipy optimizer in the future), default is `Adam`.
            max_iter (int): Max iterations, default 2000.
            verbose (bool, optional): Shows more details on optimization
                process if True.

        Todo:
            Add stochastic optimization, such as mini-batch.

        Returns:
            (np.array, value):

                losses: losses over optimization steps, (max_iter, )
                time: time taken approximately

        """
        parameters = filter(lambda p: p.requires_grad, self.parameters())

        if method == 'SGD':
            self.optimizer = torch.optim.SGD(parameters, lr=0.05, momentum=0.9)
        elif method == 'Adam':
            self.optimizer = torch.optim.Adam(parameters, lr=0.001,
                                           betas=(0.9, 0.999), eps=1e-06, weight_decay=0.0)
        elif method == 'LBFGS':
            if learning_rate is None:
                learning_rate = 1.0
            self.optimizer = torch.optim.LBFGS(parameters, lr=learning_rate,
                                            max_iter=5, max_eval=None,
                                            tolerance_grad=1e-05,
                                            tolerance_change=1e-09,
                                            history_size=50,
                                            line_search_fn=None)
        elif method == 'Adadelta':
            self.optimizer = torch.optim.Adadelta(
                parameters, lr=1.0, rho=0.9, eps=1e-06, weight_decay=0.00001)
        elif method == 'Adagrad':
            self.optimizer = torch.optim.Adagrad(
                parameters, lr=0.01, lr_decay=0, weight_decay=0)
        elif method == 'Adamax':
            self.optimizer = torch.optim.Adamax(
                parameters, lr=0.002, betas=(0.9, 0.999), eps=1e-08, weight_decay=0)
        elif method == 'ASGD':
            self.optimizer = torch.optim.ASGD(
                parameters, lr=0.01, lambd=0.0001, alpha=0.75, t0=1000000.0, weight_decay=0)
        elif method == 'RMSprop':
            self.optimizer = torch.optim.RMSprop(
                parameters, lr=0.01, alpha=0.99, eps=1e-08, weight_decay=0.00, momentum=0.01, centered=False)
        elif method == 'Rprop':
            self.optimizer = torch.optim.Rprop(
                parameters, lr=0.01, etas=(0.5, 1.2), step_sizes=(1e-06, 50))
        # scipy.optimize.minimize
        # suggest to use L-BFGS-B, BFGS
        elif method in ['CG', 'BFGS', 'Newton-CG', 'Nelder-Mead', 'Powell',
                        'L-BFGS-B', 'TNC', 'COBYLA', 'SLSQP', 'dogleg',
                        'trust-ncg']:
            print('Scipy.optimize.minimize...')
            return self._optimize_scipy(method=method, maxiter=max_iter,
                                        disp=verbose)

        else:
            raise Exception('Optimizer %s is not found. Please choose one of the'
                            'following optimizers supported in PyTorch:'
                            'Adadelt, Adagrad, Adam, Adamax, ASGD, LBFGS, '
                            'RMSprop, Rprop, SGD, LBFGS. Or the optimizers '
                            'supported scipy.optimize.minminze: BFGS, L-BFGS-B,'
                            'CG, Newton-CG, Nelder-Mead, Powell, TNC, COBYLA,'
                            'SLSQP, dogleg, trust-ncg, etc.' % method)

        losses = np.zeros(max_iter)
        tic = time()

        print('{}: Start optimizing via {}'.format(self.__class__.__name__, method))
        if verbose:
            if not method == 'LBFGS':
                for idx in range(max_iter):
                    self.optimizer.zero_grad()
                    # forward
                    loss = self.compute_loss()
                    # backward
                    loss.backward()
                    self.optimizer.step()
                    losses[idx] = loss.data.numpy()
                    print('Iter: %d\tLoss: %s' % (idx, loss.data.numpy()))

            else:
                for idx in range(max_iter):
                    def closure():
                        self.optimizer.zero_grad()
                        loss = self.compute_loss()
                        loss.backward()
                        return loss
                    loss = self.optimizer.step(closure)
                    losses[idx] = loss.data.numpy()
                    print('Iter: %d\tLoss: %s' % (idx, loss.data.numpy()))
        else:
            if not method == 'LBFGS':
                for idx in range(max_iter):
                    self.optimizer.zero_grad()
                    # forward
                    loss = self.compute_loss()
                    # backward
                    loss.backward()
                    self.optimizer.step()
                    losses[idx] = loss.data.numpy()
                    if idx % 20 == 0:
                        print('Iter: %d\tLoss: %s' % (idx, loss.data.numpy()))
            else:
                for idx in range(max_iter):
                    def closure():
                        self.optimizer.zero_grad()
                        loss = self.compute_loss()
                        loss.backward()
                        return loss
                    loss = self.optimizer.step(closure)
                    if isinstance(loss, float):  # Converged!
                        losses[idx] = loss
                        losses = losses[0:idx+1]
                        break
                    else:
                        losses[idx] = loss.data.numpy()
                    if idx % 20 == 0:
                        print('Iter: %d\tLoss: %s' % (idx, losses[idx]))
        t = time() - tic
        print('Optimization time taken: %s s' % t)
        print('Optimization method: %s' % str(self.optimizer))
        if len(losses) == max_iter:
            print('Optimization terminated by reaching the maximum iterations')
        else:
            print('Optimization terminated by getting below the tolerant error')

        return losses, t

    def _optimize_scipy(self, method='L-BFGS-B', tol=None, callback=None,
                        maxiter=1000, disp=True):
        """
        Wrapper of scipy.optimize.minimize

        Args:
             method (string): Optimization method
             maxiter (int, optional): Maximum number of iterations to perform,
                default is 1000
             disp (bool, optional): Set to True to print convergence messages.
        """
        options = dict(disp=disp, maxiter=maxiter)

        result = minimize(fun=self._loss_and_grad,
                          x0=self._get_param_array(),
                          method=method,
                          jac=True,
                          tol=tol,
                          callback=callback,
                          options=options)
        return result

    def _predict(self, input_new: TensorType, diag=True):
        # diag: is a flag indicates whether only returns the diagonal of
        # predictive variance at input_new
        # :param input_new: np.ndarray
        raise NotImplementedError

    @input_as_tensor
    def predict_f(self, input_new, diag=True):
        """
        Computes the mean and variance of the latent function at input_new
        return the diagonal of the cov matrix

        Args:
            input_new (numpy.ndarray)
        """
        return self._predict(input_new, diag=diag)

    @input_as_tensor
    def predict_y(self, input_new, diag=True):
        """
        Computes the mean and variance of observations at new inputs

        Args:
            input_new (numpy.ndarray)
        """
        mean_f, cov_f = self._predict(input_new, diag=diag)
        
        if diag:
            return self.likelihood.predict_mean_variance(mean_f, cov_f)
        else:
            return self.likelihood.predict_mean_covariance(mean_f, cov_f)

    @input_as_tensor
    def predict_f_samples(self, input_new, n_samples=1):
        """
        Return [n_samp x n_test x d_y] matrix of samples
        :param input_new:
        :param n_samples:
        :return:
        """
        mu, sigma = self.predict_f(input_new, diag=False)
        chol_s = cholesky(sigma)
        samp = mu + torch.stack([torch.mm(chol_s, TensorType(r))
                              for r in np.random.randn(n_samples, *mu.size())])
        return samp

    @input_as_tensor
    def predict_y_samples(self, input_new, n_samples=1):
        """
        Return [n_samp x n_test x d_y] matrix of samples
        :param input_new:
        :param n_samples:
        :return:
        """
        mu, sigma = self.predict_y(input_new, diag=False)
        chol_s = cholesky(sigma)
        samp = mu + torch.stack([torch.mm(chol_s, TensorType(r))
                              for r in np.random.randn(n_samples, *mu.size())])
        return samp

    # TODO: need more thought on this interface
    # convert the np operations into tensor ops
    def evaluate(self, x_test, y_test, metric='NLML'):
        """
        Evaluate the model using various metrics, including:

        - SMSE: Standardized Mean Squared Error
        - RMSE: Rooted Mean Squared Error
        - MSLL: Mean Standardized Log Loss
        - NLML: Negative Log Marginal Likelihood

        Args:
            x_test (numpy.ndarray): test input
            y_test (numpy.ndarray): test observations
            metric (string, optional): name for the metric, chosen from the list
                above.

        Returns:
            value of the metric
        """

        y_pred_mean, y_pred_var = self.predict_y(x_test)
        y_pred_mean, y_pred_var = y_pred_mean.data.numpy(), y_pred_var.data.numpy()

        if metric == 'SMSE':
            return np.power(y_pred_mean - y_test, 2).sum() / y_test.shape[0] / \
                y_test.var()
        elif metric == 'RMSE':
            return np.sqrt(np.power(y_pred_mean - y_test, 2).sum() / \
                y_test.shape[0])
        elif metric == 'MSLL':
            # single output dimension
            # predictions for each independent output dimension are the same
            # fitting training data with trivial Guassian
            y_train = self.Y.data.numpy()
            m0 = y_train.mean()
            S0 = y_train.var()
            msll = 0.5 * np.mean(np.log(2 * np.pi * y_pred_var) + \
                np.power(y_pred_mean - y_test, 2) / y_pred_var) - \
                0.5 * np.mean(np.log(2 * np.pi * S0) + \
                np.power(y_test - m0, 2) / S0)
            # 0.5 * (y_test.shape[0] * np.log(2 * np.pi * S0) + \
            # np.sum(np.power(y_test - m0, 2) / S0))
            return msll
        elif metric == 'NLML':
            return self.compute_loss().data.numpy()
        else:
            raise Exception('No such metric are supported currently, ' + 
                'select one of the following: SMSE, RSME, MSLL, NLML.')



# File: test_model.py
# File Created: Saturday, 23rd February 2019 8:30:36 pm
# Author: Steven Atkinson (steven@atkinson.mn)

"""
Tests for model.py
"""

import os, sys
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
import numpy as np
import torch as th

from gptorch.model import Param, Model, GPModel
from gptorch.kernels import Rbf
from gptorch.models import GPR


class TestParam(object):
    """
    Tests for the Param class
    """
    def test_init(self):
        # Test various permitted inits:
        Param(th.DoubleTensor([1.0]))
        Param(th.DoubleTensor([1.0]), requires_grad=False)
        Param(th.DoubleTensor([1.0]), requires_transform=False)
        Param(th.DoubleTensor([1.0]), requires_grad=False, 
            requires_transform=False)

    def test_access(self):
        """
        Test accessing the value.
        """
        p = Param(th.DoubleTensor([1.0]))
        assert isinstance(p.data, th.DoubleTensor)
        assert isinstance(p.data.numpy(), np.ndarray)

    def test_transform(self):
        """
        Test that parameters requiring a transform return the correct value.

        Currently, we obtain the untransformed variable by default.  Perhaps we
        should switch this in the future.
        """
        p = Param(th.DoubleTensor([1.0]))
        assert p.data.numpy()[0] == 1.0
        pt = Param(th.DoubleTensor([1.0]), requires_transform=True)
        assert p.transform().data.numpy()[0] == 1.0


class TestModel(object):
    """
    Tests for the Model class
    """
    pass


class TestGPModel(object):
    """
    Tests for the GPModel class
    """
    def test_predict_f_samples(self):
        # TODO mock a GPModel?  Using GPR for the moment.
        n, dx, dy = 5, 3, 2
        x, y = np.random.randn(n, dx), np.random.randn(n, dy)
        kern = Rbf(dx, ARD=True)
        gp = GPR(y, x, kern)

        n_test = 5
        x_test = np.random.randn(n_test, dx)
        f_samples = gp.predict_f_samples(x_test)
        assert isinstance(f_samples, th.Tensor)
        assert f_samples.ndimension() == 3  # [sample x n_test x dy]
        assert f_samples.shape == (1, n_test, dy)

        n_samples = 3
        f_samples_2 = gp.predict_f_samples(x_test, n_samples=n_samples)
        assert isinstance(f_samples_2, th.Tensor)
        assert f_samples_2.ndimension() == 3  # [sample x n_test x dy]
        assert f_samples_2.shape == (n_samples, n_test, dy)

    def test_predict_y_samples(self):
        # TODO mock a GPModel?  Using GPR for the moment.
        n, dx, dy = 5, 3, 2
        x, y = np.random.randn(n, dx), np.random.randn(n, dy)
        kern = Rbf(dx, ARD=True)
        gp = GPR(y, x, kern)

        n_test = 5
        x_test = np.random.randn(n_test, dx)
        y_samples = gp.predict_y_samples(x_test)
        assert isinstance(y_samples, th.Tensor)
        assert y_samples.ndimension() == 3  # [sample x n_test x dy]
        assert y_samples.shape == (1, n_test, dy)

        n_samples = 3
        y_samples_2 = gp.predict_y_samples(x_test, n_samples=n_samples)
        assert isinstance(y_samples_2, th.Tensor)
        assert y_samples_2.ndimension() == 3  # [sample x n_test x dy]
        assert y_samples_2.shape == (n_samples, n_test, dy)


"""
Kernel expectations

Supports RBF and Linear kernels

on Xcov: posterior covariance matrices for X

In Sheffield GPLVM, the multi-output GP has independent output.
Thus for each datum X[i, :], it has a diagonal covariance matrix whether
X is iid or sequential data.

But there could be the case in the future that each datum x[i, :] has a
full covariance matrix. This will be left as a future feature.
currently only support the diagonal covariance for each datum, though
preprocessing of Xcov is needed for sequential data, as follows:

Lamda: n x q
Xcov: q x n x n
    eqn (3.30) (similar thing for Xmean)
    Xcov[i, :, :] = (Kx.inverse + Lamda[:, i].diag).inverse
    
For the computation in ekernels, only the diagonal part of each 
Xcov[i, :, :] (n x n) is needed.
Xcov[i, :] <- Xcov[i].diag()
Xcov <- Xcov.t()

Not sure if those slice operations are supported by autograd, to check!

GPflow's implementation takes too much memory, but parallel.
Our current implementation is serial, small memory, could be distributed.
For n = 1000, m = 50, q = 10. GPtorch takes 2s to forward, and 1.5s backward.
Comparable to GPflow at this scale.
"""

# Xcov.transform() --> Xcov

# July 20
# add more parallel implementation (more memory consumption, but faster)
# the covariance matrix can be full matrix, not diagonal


from __future__ import absolute_import
import gptorch.kernels
import torch as th
from torch.autograd import Variable
from gptorch.functions import cholesky, trtrs

float_type = th.DoubleTensor

# TODO: necesary to be able to handle any type of kernels (BB)

class Rbf(gptorch.kernels.Rbf):
    def eKxx(self, X):
        """psi_0 in GPy.

        Args:
            X (np.ndarray or Variable): input points of GP

        Returns:
            (Variable): an array with elements are all the kernel variance
        """

        return self.Kdiag(X)

    def eKxz(self, Z, Xmean, Xcov, requires_transform=True):
        """Computes the expectation, eKzxKxz_1, n x m:

        .. math::
            <K_{xz}>_{q(X)}

        Closed form solution for RBF kernel expectation is in the appendix B.2
        of Damianou's Diss.

        Args:
            Z (Variable): m x q inducing input
            Xmean (Variable): n x q mean of input X
            Xcov (Varible): n x q posterior covariance of X, for both iid and
            seq
            
        """
        num_data = Xmean.size(0)
        dim_latent = Xmean.size(1)
        num_inducing = Z.size(0)
        length_scales = self.length_scales.transform()
        variance = self.variance.transform()

        if requires_transform:
            Xcov = Xcov.transform()

        # compute scaled squared distance matrix
        Xmean = Xmean / th.sqrt(length_scales.pow(2).expand_as(Xmean) + Xcov)
        # every row of Xmean_2 is the repetition of the Euclidean norm of
        # Xmean[i, :], n x m
        Xmean_2 = Xmean.pow(2).sum(1).view(-1, 1).expand(num_data, num_inducing)
        # compute the eKxz[i, :] by iterating over each datum, 
        # could be distributed
        eKxz_list = []
        for i in xrange(num_data):
            Zi = Z / \
                 th.sqrt(length_scales.pow(2).expand_as(Z) +
                         Xcov[i, :].expand_as(Z))
            Xmean_Zi = Xmean[i, :].unsqueeze(0).mm(Zi.t())
            # Zi_2 = Zi.pow(2).sum(1).t()
            Zi_2 = Zi.pow(2).sum(1).view(1, -1)
            log_term = (Xcov[i, :] / length_scales.pow(2) +
                        Variable(th.ones(dim_latent).type(float_type))).log().\
                sum().expand_as(Xmean_Zi)
            eKxz_list.append(Xmean_2[i, :].unsqueeze(0) + Zi_2 - \
                             2 * Xmean_Zi + log_term)

        return th.exp(-0.5 * th.cat(eKxz_list, 0)) * \
            variance.expand(num_data, num_inducing)


    def eKxz_parallel(self, Z, Xmean, Xcov):
        # TODO: add test
        """Parallel implementation (needs more space, but less time)
        Refer to GPflow implementation

        Args:
            Args:
            Z (Variable): m x q inducing input
            Xmean (Variable): n x q mean of input X
            Xcov (Varible): posterior covariance of X
                two sizes are accepted:
                    n x q x q: each q(x_i) has full covariance
                    n x q: each q(x_i) has diagonal covariance (uncorrelated),
                        stored in each row
        Returns:
            (Variable): n x m
        """

        # Revisit later, check for backward support for n-D tensor
        n = Xmean.size(0)
        q = Xmean.size(1)
        m = Z.size(0)
        if Xcov.dim() == 2:
            # from flattered diagonal to full matrix
            cov = Variable(th.Tensor(n, q, q).type(float_type))
            for i in range(Xmean.size(0)):
                cov[i] = Xcov[i].diag()
            Xcov = cov
            del cov
        length_scales = self.length_scales.transform()
        Lambda = length_scales.pow(2).diag().unsqueeze(0).expand_as(Xcov)
        L = cholesky(Lambda + Xcov)
        xz = Xmean.unsqueeze(2).expand(n, q, m) - Z.unsqueeze(0).expand(n, q, m)
        Lxz = th.triangular_solve(xz, L, upper=False)[0]
        half_log_dets = L.diag().log().sum(1) \
                        - length_scales.log().sum().expand(n)

        return self.variance.transform().expand(n, m) \
               * th.exp(-0.5 * Lxz.pow(2).sum(1) - half_log_dets.expand(n, m))


    def eKzxKxz(self, Z, Xmean, Xcov, requires_transform=True, sum=True):
        """Computes the expectation, psi_2, m x m:

        .. math::
            <K_{zx}L_{xz}>_{q(X)}

        Args:
            Z (Variable): m x q inducing input
            Xmean (Variable): n x q mean of input X
            Xcov (Varible): n x q posterior covariance of X, for both iid and seq
            requires_transform (bool, optional): True if ``Xcov`` needs transform
            sum (bool, optional): False if returning a list of (Psi_2)_i,
                True if returning the sum of (Psi_2)_i, i=1,..,n
        """
        num_data = Xmean.size(0)
        dim_latent = Xmean.size(1)
        length_scales = self.length_scales.transform()
        variance = self.variance.transform()
        if requires_transform:
            Xcov = Xcov.transform()

        # compute shared term, for all i
        eKzxKxz = Variable(th.zeros(Z.size(0), Z.size(0)).type(float_type))
        if not sum:
            eKzxKxz_list = []
        # Z1 is only scaled by length_scales, as the first term of the exponent
        Z1 = Z / length_scales.expand_as(Z)
        ZZT = Z1.mm(Z1.t())
        Z1_2 = Z1.pow(2).sum(1).expand_as(ZZT)
        shared_term = Z1_2 + Z1_2.t() - 2 * ZZT
        # \Xmean_{i,j}
        Xmean = Xmean / \
             th.sqrt(length_scales.pow(2).expand_as(Xmean) +
                     2 * Xcov)
        Xmean_2 = Xmean.pow(2).sum(1).view(-1, 1)

        for i in xrange(num_data):
            Zi = Z / \
                 th.sqrt(length_scales.pow(2).expand_as(Z) +
                         2 * Xcov[i, :].expand_as(Z))
            Zi_2 = Zi.pow(2).sum(1).expand_as(eKzxKxz)

            Xmean_Zi = Zi.mm(Xmean[i, :].unsqueeze(1)).expand_as(eKzxKxz) + \
                    Xmean[i, :].unsqueeze(0).mm(Zi.t()).expand_as(eKzxKxz)

            log_term = th.log(Xcov[i, :] * 2 / length_scales.pow(2) +
                        Variable(th.ones(dim_latent).type(float_type))).sum().\
                expand_as(eKzxKxz)

            eKzxKxz_i = \
                th.exp(-0.25 * shared_term - Xmean_2[i, :].expand_as(eKzxKxz)
                       - 0.25 * (Zi_2 + Zi_2.t() + 2 * Zi.mm(Zi.t())) + Xmean_Zi
                       - 0.5 * log_term) * variance.pow(2).expand_as(eKzxKxz)
            if sum:
                eKzxKxz += eKzxKxz_i
            else:
                eKzxKxz_list.append(eKzxKxz_i)

        return eKzxKxz if sum else eKzxKxz_list


class Linear(gptorch.kernels.Rbf):
    pass



# Steven Atkinson
# satkinso@nd.edu
# 

"""
Tests on densities:
"""

from __future__ import absolute_import
from gptorch.densities import GaussianMultivariateDiagonal, \
    GaussianMultivariateFull
from gptorch import util
from torch.autograd import Variable
import torch
import numpy as np
from unittest import TestCase
import os

# For computations that are limited by computational precision
comp_tol = 1.0e-12
# For comptutaions that are limited by Monte Carlo convergence (sampling tests)
# This is always dicey, so the tolerance is pretty relaxed
mc_tol = 0.2


def _get_matrix(s):
    x_np = np.loadtxt(os.path.join(os.path.dirname(__file__), 'data', 
        'densities', s + '.dat'))
    if x_np.size == 1:
        x_np = x_np.reshape((1))
    return util.as_variable(x_np)


def _err(measured, target, relative=True):
    err = torch.abs(measured - target)
    if relative:
        err /= target
    return torch.max(err).data.numpy()


class TestGaussianMultivariateDiagonal(TestCase):
    @staticmethod
    def _get_matrix(s):
        return _get_matrix(os.path.join('MultivariateGaussianDiagonal/', s))

    @staticmethod
    def _get_mu_sigma():
        return TestGaussianMultivariateDiagonal._get_matrix('mu'), \
               TestGaussianMultivariateDiagonal._get_matrix('sigma')

    def test_pdf(self):
        mu, sigma = TestGaussianMultivariateDiagonal._get_mu_sigma()
        x = TestGaussianMultivariateDiagonal._get_matrix('x')
        pdf_true = TestGaussianMultivariateDiagonal._get_matrix('p_of_x')
        d = GaussianMultivariateDiagonal(mu, sigma)
        pdf_computed = d.pdf(x)
        assert isinstance(pdf_computed, Variable)
        assert len(pdf_computed.size()) == 1  # Return a 1D array
        err = _err(pdf_computed, pdf_true)
        assert err < comp_tol

    def test_sample(self):
        mu, sigma = TestGaussianMultivariateDiagonal._get_mu_sigma()
        d = GaussianMultivariateDiagonal(mu, sigma)
        s = d.sample(10 ** 7)
        mu_measured = torch.mean(s, 0)
        cov_measured = util.as_variable(np.cov(s.data.numpy().transpose()))
        mu_err = _err(mu_measured, mu)
        cov_err = _err(cov_measured, sigma.diag(), False)
        print(mu_err)
        print(cov_err)
        assert mu_err < mc_tol
        assert cov_err < mc_tol


class TestGaussianMultivariateFull(TestCase):
    @staticmethod
    def _get_matrix(s):
        return _get_matrix('MultivariateGaussianFull/' + s)

    @staticmethod
    def _get_mu_sigma():
        return TestGaussianMultivariateFull._get_matrix('mu'), \
               TestGaussianMultivariateFull._get_matrix('sigma')

    def test__get_precision(self):
        mu, sigma = TestGaussianMultivariateFull._get_mu_sigma()
        sigma_inv = TestGaussianMultivariateFull._get_matrix('sigma_inv')
        d = GaussianMultivariateFull(mu, sigma, compute_precision=True)
        precision_err = _err(d._precision, sigma_inv)
        assert precision_err < comp_tol

    def test__get_chol(self):
        mu, sigma = TestGaussianMultivariateFull._get_mu_sigma()
        sigma_chol = TestGaussianMultivariateFull._get_matrix('sigma_cholesky')
        d = GaussianMultivariateFull(mu, sigma, compute_chol=True)
        chol_err = _err(d._l, sigma_chol, False)
        assert chol_err < comp_tol

    def test__get_coefficient(self):
        mu, sigma = TestGaussianMultivariateFull._get_mu_sigma()
        d = GaussianMultivariateFull(mu, sigma)
        d._get_coefficient()
        coefficient_true = TestGaussianMultivariateFull._get_matrix(
            'coefficient')
        coeff_err = _err(d._coefficient, coefficient_true)
        assert coeff_err < comp_tol

    def test_pdf(self):
        mu, sigma = TestGaussianMultivariateFull._get_mu_sigma()
        x = TestGaussianMultivariateFull._get_matrix('x')
        pdf_true = TestGaussianMultivariateFull._get_matrix('p_of_x')
        d = GaussianMultivariateFull(mu, sigma)
        pdf = d.pdf(x)
        pdf_err = _err(pdf, pdf_true)
        assert pdf_err < comp_tol

    def test_sample(self):
        mu, sigma = TestGaussianMultivariateFull._get_mu_sigma()
        d = GaussianMultivariateFull(mu, sigma)
        s = d.sample(10 ** 7)
        mu_measured = torch.mean(s, 0)
        cov_measured = util.as_variable(np.cov(s.data.numpy().transpose()))
        mu_err = _err(mu_measured, mu)
        cov_err = _err(cov_measured, sigma)
        print(mu_err)
        print(cov_err)
        assert mu_err < mc_tol
        assert cov_err < mc_tol




"""
``gptorch.functions`` contains self-defined functions subclassing
:modile:`torch.autograd.Function` that performs Cholesky decomposition, solving
linear triangular equations, and misc functions.
"""

from __future__ import absolute_import
from .util import as_tensor
import torch
from torch.nn import functional as F
import numpy as np

torch_version = [int(s) for s in torch.__version__.split(".")]
_potri = torch.cholesky_inverse if torch_version >= [1, 1, 0] else torch.potri
    

def jit_op(op, x: torch.Tensor, max_tries: int=10, verbose: bool=False) \
        -> torch.Tensor:
    """
    Attempt a potentially-unstable linear algebra operation on a matrix.
    If it fails, then try adding more and more jitter and try again...
    """
    jitter_size = x.diag().mean()
    try:
        return op(x)
    except Exception as e:
        if verbose:
            print("Op {} failed (initial try)".format(op.__name__))

    for i in range(max_tries):
        try:
            this_jitter = 10.0 ** (-max_tries + i) * torch.eye(*x.shape, 
                dtype=x.dtype)
            return op(x + this_jitter)
        except RuntimeError as e:
            if verbose:
                print("Op {} failed (try {} / {})".format(op.__name__, i + 1, 
                    max_tries))
    raise RuntimeError("Max tries exceeded.")


def cholesky(x: torch.Tensor) -> torch.Tensor:
    return jit_op(torch.cholesky, x)


def cholesky_inverse(x: torch.Tensor, upper=False) -> torch.Tensor:
    """
    Inverse of a matrix based on its Cholesky.
    """
    return _potri(x, upper=upper)


def inverse(x: torch.Tensor) -> torch.Tensor:
    return jit_op(torch.inverse, x)


def lt_log_determinant(L):
    """
    Log-determinant of a triangular matrix

    Args:
        L (Variable): Lower-triangular matrix to take log-determinant of.
    """
    return L.diag().log().sum()


def SoftplusInv(y, lower=1e-6):
    '''Transforms for handling constraints on parameters, e.g. positive variance
    For get the initial value of x, where

    .. math::
        y = \mathrm{Softplus}(x) = \log(1 + e^x)

    SoftplusInv is used to represent the positive constraints of some
    parameters, such as variance.

    Args:
        y (numpy.ndarray or real number): output of softplus,
            value of the parameter value

    Returns:
        the 'free' parameter used in optimization
    '''
    x = torch.log(torch.exp(y - lower) - 1.)
    if y.numpy().any() > 35:
        return y - lower
    else:
        return x


def transform(variable):
    # intent to replace the transform method within the Param class
    assert isinstance(variable, Variable), "Input to this function should be a Variable"
    if variable.requires_transform:
        return F.softplus(variable, threshold=35)
    else:
        return variable


def trtrs(b: torch.Tensor, a: torch.Tensor, lower=True) -> torch.Tensor:
    return torch.triangular_solve(b, a, upper=not lower)[0]



#
# Apr 15, 2017  Yinhao Zhu
#
# Class for GP regression

"""
Class for Vanilla GP regression.
"""

from __future__ import absolute_import
import torch
import numpy as np

from .. import kernels
from ..model import GPModel, Param
from .. import likelihoods
from ..functions import cholesky, inverse, lt_log_determinant, trtrs
from ..util import TensorType


class GPR(GPModel):
    """
    Gaussian Process Regression
    """
    def __init__(self, observations, input, kernel, mean_function=None,
                 likelihood=None, name='gpr'):
        """
        Default likelihood is Gaussain, mean function is zero.

        Args:
            observations (np.ndarray or TensorType): a matrix of observed outputs.
                Each datum is a row in the matrix (# rows = # training data,
                # columns = output dimensionality)
            input (np.ndarray or TensorType): the observed inputs.
                If input is a numpy matrix or torch TensorType, then each datum is
                a row in the matrix (# rows = # training data,
                # columns = input dimensionality).
            kernel (Kernel): The kernel function for computing the covariance
                matrices
            mean_function (MeanFunction): class for populating the design matrix
                for computing the parameterized mean function.
            likelihood (Likelihood): A likelihood model
        """
        if likelihood is None:
            likelihood = likelihoods.Gaussian()
        super().__init__(observations, input, kernel, likelihood,
                                  mean_function, name)

    def compute_loss(self):
        """
        Loss is equal to the negative of the log likelihood

        Adapted from Rasmussen & Williams, GPML (2006), p. 19, Algorithm 2.1.
        """

        num_input = self.Y.size(0)
        dim_output = self.Y.size(1)

        L = cholesky(self._compute_kyy())
        alpha = trtrs(self.Y, L)
        const = TensorType([-0.5 * dim_output * num_input * np.log(2 * np.pi)])
        loss = 0.5 * alpha.pow(2).sum() + dim_output * lt_log_determinant(L) \
            - const
        return loss

    def _compute_kyy(self):
        """
        Computes the covariance matrix over the training inputs

        Returns:
            Ky (TensorType)
        """
        num_input = self.Y.size(0)

        return self.kernel.K(self.X) + \
        (self.likelihood.variance.transform()).expand(
            num_input, num_input).diag().diag()

    def _predict(self, input_new: TensorType, diag=True):
        """
        This method computes

        .. math::
            p(F^* | Y )

        where F* are points on the GP at input_new, Y are observations at the
        input X of the training data.
        :param input_new: test inputs; should be two-dimensional
        """
        k_ys = self.kernel.K(self.X, input_new)

        L = cholesky(self._compute_kyy())
        A = trtrs(k_ys, L)
        V = trtrs(self.Y, L)
        mean_f = A.t() @ V

        if self.mean_function is not None:
            mean_f += self.mean_function(input_new)

        var_f_1 = self.kernel.Kdiag(input_new) if diag else \
            self.kernel.K(input_new)  # Kss

        if diag:
            var_f_2 = (A * A).sum(0)
        else:
            var_f_2 = A.t() @  A
        var_f = var_f_1 - var_f_2

        return mean_f, var_f




"""
Basic tests for the repo
"""


def test_torch_dtype():
    """
    Ensure that importing gptorch does not change the default dtype for torch.
    """

    import torch
    dtype = torch.get_default_dtype()
    import gptorch
    new_dtype = torch.get_default_dtype()

    assert new_dtype == dtype



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import sys

valid_python = sys.version_info[0] >= 3 and sys.version_info[1] >= 5
if not valid_python:
    raise RuntimeError("gptorch requires python 3.5+")

# Python 3.5 can't do matplotlib 3.1 and up.
matplotlib_dependency = "matplotlib>=2.1.2" if sys.version_info[1] >= 6 else \
    "matplotlib>=2.1.2, <3.1"

requirements = [
    "numpy>=1.10",
    "scipy>=0.18",
    matplotlib_dependency,
    "scikit-learn>=0.19.1",
    "torch",  # conda install pytorch -c pytorch
    "pytest>=3.5.0",
    "graphviz>=0.9",
    "jupyter"
]

setup(name="gptorch",
    version="0.2.2",
    description="gptorch - a Gaussian process toolbox built on PyTorch",
    author="Yinhao Zhu, Steven Atkinson",
    author_email="yzhu10@nd.edu, satkinso@nd.edu",
    url="https://github.com/cics-nd/gptorch",
    install_requires=requirements,
    packages=find_packages(),
    )



#
# Apr 17, 2017    Yinhao Zhu
#
# class for likelihoods: p(y | f)

from __future__ import absolute_import
from . import densities
from .model import Model, Param

from torch.nn import Parameter
import torch as th

# from functions import SoftplusInv
# from torch.nn import functional as F

float_type = th.DoubleTensor


class Likelihood(Model):

    def __init__(self):
        super(Likelihood, self).__init__()

    def predict_mean_variance(self, mean_f, var_f):
        """
        The likelihood of observations y given latent functions f is p(y | f)
        It computes the mean and variance of observations Y given latent
        functions F, which itself is Gaussian.

        p(y) = \int p(y | f) p(f) df

        Gauss-Hermite quadrature is used for numerical integration.
        :param mean_f: mean of latent functions
        :param var_f: variance of the latent functions
        :return: mean and variance of the observations
        """
        # TODO: Gauss-Hermite quadrature
        raise NotImplementedError

    def forward(self):
        return None

    # def __repr__(self):
    #     tmpstr = self.__class__.__name__ + ' (\n'
    #     for name, param in self._parameters.items():
    #         tmpstr = tmpstr + name + str(param.data) + '\n'
    #     tmpstr = tmpstr + ')'
    #     return tmpstr


class Gaussian(Likelihood):
    # Possibly replace these with torch.distributions?
    def __init__(self, variance=1.0):
        super(Gaussian, self).__init__()
        self.variance = Param(th.Tensor([variance]).type(float_type),
                              requires_transform=True)

    def logp(self, F, Y):
        """
        Evaluate the log-density of targets at Y, given a Gaussian density 
        centered at F.

        :param F: Center of the density
        :type F: torch.autograd.Variable
        :param Y: Targets where we want to compute the log-pdf
        :type Y: torch.autograd.Variable
        """
        return densities.gaussian(F, Y, self.variance.transform())

    def predict_mean_variance(self, mean_f, var_f):
        """
        Integrate the input (a Gaussian with provided mean & variance) over the
        likelihood density

        :param mean_f: Mean of input Gaussian
        :type mean_f: torch.autograd.Variable
        :param var_f: Variance of input Gaussian
        :type var_f: torch.autograd.Variable

        :return: (torch.autograd.Variable, torch.autograd.Variable) mean & var
        """
        # TODO: consider mulit-output case
        # stupid syntax - expecting broadcasting in PyTorch
        return mean_f, var_f + self.variance.transform().expand_as(var_f)

    def predict_mean_covariance(self, mean_f, cov_f):
        return mean_f, cov_f + self.variance.transform().expand_as(cov_f).\
            diag().diag()



"""
Gaussian Process Latent Variable Model (GPLVM)

"""
#
# Yinhao Zhu, May, July 2017
#
#

from __future__ import  absolute_import
from gptorch.model import GPModel, Param
from gptorch.likelihoods import Gaussian
from gptorch.mean_functions import Zero
from gptorch.functions import cholesky, trtrs
from gptorch import ekernels, kernels
from gptorch import util

import numpy as np
import torch as th
from torch.autograd import Variable
from sklearn.decomposition import PCA
from time import time
try:
    # Python 2
    from future_builtins import filter
except ImportError:
    # Python 3
    pass

float_type = th.DoubleTensor


class GPLVM(GPModel):
    """
    Variational GPLVM

    Reference:
        Damianou, Andreas. Deep Gaussian processes and variational propagation
        of uncertainty. Diss. University of Sheffield, 2015.

    """
    def __init__(self, observations, dim_latent, num_inducing, Xmean=None,
                 inducing_points=None, kernel=None, kernel_x=None,
                 data_type='iid', collapsed_bound=True, large_p=False):
        """
        Initialization for the variational GPLVM

        Args:
            observations (np.ndarray): Observed data for unsupervised learning
            dim_latent (int): Dimensionality of the latent variables
            num_inducing (int): Number of inducing points
            Xmean (np.ndarray): Latent variable means (if None, will be init by
                PCA)
            inducing_points (np.ndarray): Inducing points, Z
            kernel (gptorch.Kernel):
            data_type (string): ``iid`` or ``seq`` (sequential)
            collapsed_bound (bool, optional): True for computing the ELBO when
                inducing variables are collapsed (second bound), False for
                the uncollapsed bound
            large_p (bool, optional): True for the case of small n, large p
                (HD video), False for the case of large p, small n.
                This option affects the computation of KL(q(X) || p(X))
        """
        assert isinstance(observations, np.ndarray), \
            "Observation matrix should be a np.ndarray."

        if Xmean is None:
            print("GPLVM: Initialize the Xmean using PCA")
            if large_p:
                pca_sklean = PCA(n_components=dim_latent)
                Xmean = pca_sklean.fit_transform(observations)
            else:
                Xmean = util.as_variable(util.PCA(observations, dim_latent))
        else:
            assert isinstance(Xmean, np.ndarray), \
                "Initialization of posterior mean of latent variables should" \
                " be np.ndarray."

        if kernel is None:
            kernel = ekernels.Rbf(dim_latent, ARD=True)
        else:
            assert dim_latent == kernel.input_dim, \
                "Input dimensionality of kernel must be equal to dim_latent."
            assert isinstance(kernel, ekernels.Rbf), \
                "Supports only ekernel.Rbf currently."

        super(GPLVM, self).__init__(observations, Xmean, kernel,
                                    Gaussian(), Zero(), name='GPLVM')
        del self.X

        # flag to distinguish training and testing mode
        self.inference = False
        self.data_type = data_type
        self.is_collapsed = collapsed_bound
        self.is_large_p = large_p

        # Setup for test time inference
        # test data will be assigned in the projection method
        self.Y_test = None
        # latent variable mean and covariance for the test data
        self.Xmean_test = None
        self.Xcov_test = None
        # observed dimensions of the test data
        self.observed_dims = None
        self.saved_terms = {}
        if self.is_large_p:
            # saved for faster computation in the lower bound, n x n
            self.saved_terms['YYT'] = self.Y.mm(self.Y.t())

        if self.data_type == 'iid':
            # posterior mean of X initialized by PCA of Y: n x q
            # self.Xmean = Param(th.from_numpy(Xmean).type(float_type))
            self.Xmean = Param(Xmean.data)
            # posterior covariance of X: n x q
            self.Xcov = Param(0.5 * th.ones(self.Xmean.size()).type(float_type)
                              + 0.001 * th.randn(self.Xmean.size()).type(float_type),
                              requires_transform=True)
        else:
            # sequential data
            # temporal kernel for the GP from time t to latent variables X
            if isinstance(kernel_x, kernels.Kernel):
                assert kernel_x.input_dim == 1, \
                    "Currently only supports time input, i.e. kernel with " \
                    "one dimension input"
                self.kernel_x = kernel_x
            else:
                # TODO: kernel_x, better initialization needed!
                self.kernel_x = kernels.Rbf(1, variance=0.5, length_scales=0.5)
                self.kernel_x.variance.requires_grad = False

            # 1) vanilla, O(n^2*q) parameters for q(X) (not scalable)
            # ----- 2) Reparameterization (3.30) p58 in Damianou Diss. ------ current impl
            # 3) recognition model (will be the useful one) (RNN)

            # TODO: add the ability to handle multiple sequences (regressive)

            # 2) Reparameterization (3.30) p58 in Damianou Diss.
            # posterior mean of X initialized by PCA of Y: n x q
            # Xmean = th.Tensor(x_post_mean).type(float_type)
            # intermediate variables, useful for inference purpose, or other queries
            self.Xmean = Variable(Xmean)
            # init the cov matrix by using the kernel
            # timestamp is not required for stationary kernels
            Kx = self.kernel_x.K(np.array(xrange(self.Y.size(0)))[:, None])
            # optimization parameters are mu_bar as in (3.30)
            self.Xmean_bar = Param(Kx.data.inverse().mm(Xmean))

            # assume the posterior S is the same as the prior Kx
            # self.lambda_ = Param(th.zeros(Xmean.size()).type(float_type))
            # assume the posterior S is close to the prior Kx
            # Constrain the Lambda to be positive, to ensure the S is PSD
            self.Lambda = Param(th.rand(Xmean.size()).type(float_type) * 0.25,
                                requires_transform=True)
            # dummy initialization, n x q
            self.Xcov = Variable(th.ones(Xmean.size()).type(float_type) * 0.5)

        if inducing_points is not None:
            if isinstance(inducing_points, np.ndarray):
                assert inducing_points.shape[0] == num_inducing and \
                       inducing_points.shape[1] == dim_latent, \
                    "Dimensionality of inducing points does not match"
                self.Z = Param(th.from_numpy(inducing_points).type(float_type))
        else:
            # inducing points Z, init with subset of posterior mean of X
            z_np = Xmean.data.numpy()[np.random.choice(Xmean.size(0),
                                                       num_inducing,
                                                       replace=False)]
            self.Z = Param(float_type(z_np))


        # Uncollapsed case, the number of parameters associated with inducing points
        #  variance is O(m^2)
        # TODO: stochasitic optimization with the uncollpased bound
        # MNIST data set 60k, 28 x 28 digits
        if not self.is_collapsed:
            # posterior mean of inducing variables U, init with subset of observations
            self.Umean = Param(th.from_numpy(self.Y[np.random.choice(
                self.Y.size(0), num_inducing, replace=False)]).type(float_type))
            # posterior variance of inducing variables U: m x m
            # needs parameterization of cov matrix, e.g. Chol decomposition
            self.Ucov = Param(0.5 * th.ones(num_inducing, num_inducing),
                               requires_transform=True)

        # self.jitter = Param(th.FloatTensor([1e-4]), requires_transform=True)
        self.jitter = Variable(th.Tensor([1e-6]).type(float_type))

        # computes the total number of parameters to optimize over
        num_parameters = 0
        for param in self.parameters():
            num_parameters += param.data.numpy().size
        print('GPLVM: Number of optimization parameters is  %d' % num_parameters)

    def compute_loss(self):
        """
        Computation graph for the ELBO (Evidence Lower Bound) of
        the variational GPLVM
        For the implementation details, please see ``notes/impl_gplvm``.

        """
        num_data = self.Y.size(0)
        dim_output = self.Y.size(1)
        dim_latent = self.Z.size(1)
        num_inducing = self.Z.size(0)

        var_kernel = self.kernel.variance.transform()
        var_noise = self.likelihood.variance.transform()

        # computes kernel expectations
        eKxx = num_data * var_kernel
        if self.data_type == "iid":
            eKxz = self.kernel.eKxz(self.Z, self.Xmean, self.Xcov)
            eKzxKxz = self.kernel.eKzxKxz(self.Z, self.Xmean, self.Xcov)
        else:
            # seq data
            # compute S_j's and mu_bar_j's (reparameterization: forward)
            # self.Xmean, self.Xcov = self._reparam_vargp(self.Xmean_bar, self.Lambda)
            Kx = self.kernel_x.K(np.array(xrange(self.Y.size(0)))[:, None])
            # print(Kx.data.eig())
            Lkx = cholesky(Kx, flag='Lkx')
            # Kx_inverse = inverse(Kx)
            self.Xmean = Kx.mm(self.Xmean_bar)
            Xcov = []
            # S = []
            Le = []
            In = Variable(th.eye(num_data).type(float_type))
            for j in xrange(dim_latent):
                Ej = Lkx.t().mm(self.Lambda.transform()[:, j].diag()).mm(Lkx) + In
                # print(Ej.data.eig())
                Lej = cholesky(Ej, flag='Lej')
                Lsj = trtrs(Lej, Lkx.t()).t()
                Sj = Lsj.mm(Lsj.t())
                Xcov.append(Sj.diag().unsqueeze(1))
                # S.append(Sj)
                Le.append(Lej)
            self.Xcov = th.cat(Xcov, 1)
            eKxz = self.kernel.eKxz(self.Z, self.Xmean, self.Xcov, False)
            eKzxKxz = self.kernel.eKzxKxz(self.Z, self.Xmean, self.Xcov, False)

        # compute ELBO
        # add jitter
        # broadcast update
        Kzz = self.kernel.K(self.Z) + self.jitter.expand(self.Z.size(0)).diag()
        L = cholesky(Kzz, flag='Lkz')
        A = trtrs(L, trtrs(L, eKzxKxz).t()) / var_noise.expand_as(L)
        B = A + Variable(th.eye(num_inducing).type(float_type))
        LB = cholesky(B, flag='LB')

        # log|B|
        # log_det_b = LB.diag().log().sum()

        log_2pi = Variable(th.Tensor([np.log(2 * np.pi)]).type(float_type))
        elbo = - dim_output * (LB.diag().log().sum() +
                               0.5 * num_data * (var_noise.log() + log_2pi))
        elbo -= 0.5 * dim_output * (eKxx / var_noise - A.trace())

        if not self.is_large_p:
            # distributed
            # C = Variable(th.zeros(num_inducing, dim_output))
            # for i in xrange(num_data):
            #     C += Psi[i, :].unsqueeze(1).mm(self.Y[i, :].unsqueeze(0))
            C = eKxz.t().mm(self.Y)
            D = trtrs(LB, trtrs(L, C))
            elbo -= 0.5 * (
                self.Y.t().mm(self.Y) / var_noise.expand(dim_output, dim_output)
                - D.t().mm(D) / var_noise.pow(2).expand(dim_output,
                                                        dim_output)).trace()
        else:
            # small n, pre-compute YY'
            # YYT = self.Y.mm(self.Y.t())
            D = trtrs(LB, trtrs(L, eKxz.t()))
            W = Variable(th.eye(num_data).type(float_type)) \
                / var_noise.expand(num_data, num_data) \
                - D.t().mm(D) / var_noise.pow(2).expand(num_data, num_data)
            elbo -= 0.5 * (W.mm(self.saved_terms['YYT'])).trace()

        # KL Divergence (KLD) btw the posterior and the prior
        if self.data_type == 'iid':
            const_nq = Variable(th.Tensor([num_data * dim_latent]).type(float_type))
            # eqn (3.28) below p57 Damianou's Diss.
            KLD = 0.5 * (self.Xmean.pow(2).sum() + self.Xcov.transform().sum() -
                         self.Xcov.transform().log().sum() - const_nq)
        else:
            # seq data (3.29) p58
            # Xmean n x q
            # S: q x n x n
            # Kx, Kx_inverse
            KLD = Variable(th.Tensor([-0.5 * num_data * dim_latent]).type(float_type))
            KLD += 0.5 * self.Xmean_bar.mm(self.Xmean_bar.t()).mm(Kx.t()).trace()
            for j in xrange(dim_latent):
                Lej_inv = trtrs(Le[j], In)
                KLD += 0.5 * Lej_inv.t().mm(Lej_inv).trace() + Le[j].diag().log().sum()

        elbo -= KLD
        return -elbo

    def _pre_compute(self):
        """Pre-computation for the projection

        Fixed terms in test time are manually identified,
        Only iid latent is implemented.
        """
        # Save the fixed terms here
        # self.saved_terms = {}
        if self.observed_dims is not None:
            # select observed dims to compute
            Y = th.cat((self.Y.index_select(1, self.observed_dims),
                        self.Y_test), 0)
            self.saved_terms['YYT'] = Y.mm(Y.t())

        # computes kernel expectations
        if self.data_type == "iid":
            eKxz = self.kernel.eKxz(self.Z, self.Xmean, self.Xcov)
            eKzxKxz = self.kernel.eKzxKxz(self.Z, self.Xmean, self.Xcov)
            self.saved_terms['eKxz'] = eKxz
            self.saved_terms['eKzxKxz'] = eKzxKxz
        else:
            print("regressive case, not implemented")

        Kzz = self.kernel.K(self.Z) + self.jitter.expand(self.Z.size(0)).diag()
        L = cholesky(Kzz, flag='L')
        self.saved_terms['L'] = L

    def _compute_loss_inference(self):
        """Computes the loss in the inference mode, e.g. for projection.
        Handles both fully observed and partially observed data.

        Only iid latent is implemented.
        """
        num_data_train = self.Y.size(0)
        # dim_output_train = self.Y.size(1)
        dim_latent = self.Z.size(1)
        num_inducing = self.Z.size(0)
        num_data_test = self.Y_test.size(0)
        # total number of data for inference
        num_data = num_data_train + num_data_test
        # dimension of output in the test time
        dim_output = self.Y_test.size(1)
        # whole data for inference
        if self.observed_dims is None:
            Y = th.cat((self.Y, self.Y_test), 0)
        else:
            Y = th.cat((self.Y.index_select(1, self.observed_dims),
                        self.Y_test), 0)

        var_kernel = self.kernel.variance.transform()
        var_noise = self.likelihood.variance.transform()

        # computes kernel expectations
        # eKxx = num_data * self.kernel.eKxx(self.Xmean).sum()
        eKxx = num_data * var_kernel
        if self.data_type == "iid":
            eKxz_test = self.kernel.eKxz(self.Z, self.Xmean_test, self.Xcov_test)
            eKzxKxz_test = self.kernel.eKzxKxz(self.Z, self.Xmean_test, self.Xcov_test)
            eKxz = th.cat((self.saved_terms['eKxz'], eKxz_test), 0)
            eKzxKxz = self.saved_terms['eKzxKxz'] + eKzxKxz_test
        else:
            print("regressive case not implemented")

        # compute ELBO
        L = self.saved_terms['L']
        A = trtrs(L, trtrs(L, eKzxKxz).t()) / var_noise.expand_as(L)
        B = A + Variable(th.eye(num_inducing).type(float_type))
        LB = cholesky(B, flag='LB')

        log_2pi = Variable(th.Tensor([np.log(2 * np.pi)]).type(float_type))
        elbo = - dim_output * (LB.diag().log().sum() +
                               0.5 * num_data * (var_noise.log() + log_2pi))
        elbo -= 0.5 * dim_output * (eKxx / var_noise - A.diag().sum())

        if not self.is_large_p:
            # distributed
            # C = Variable(th.zeros(num_inducing, dim_output))
            # for i in xrange(num_data):
            #     C += Psi[i, :].unsqueeze(1).mm(self.Y[i, :].unsqueeze(0))
            C = eKxz.t().mm(Y)
            D = trtrs(LB, trtrs(L, C))
            elbo -= 0.5 * (
                Y.t().mm(Y) / var_noise.expand(dim_output, dim_output)
                - D.t().mm(D) / var_noise.pow(2).expand(dim_output,
                                                        dim_output)).trace()
        else:
            # small n, pre-compute YY'
            # YYT = self.Y.mm(self.Y.t())
            D = trtrs(LB, trtrs(L, eKxz.t()))
            W = Variable(th.eye(num_data).type(float_type)) / var_noise.expand(
                num_data, num_data) - \
                D.t().mm(D) / var_noise.pow(2).expand(num_data, num_data)
            elbo -= 0.5 * (W.mm(self.saved_terms['YYT'])).trace()

        # KL Divergence (KLD) btw the posterior and the prior
        if self.data_type == 'iid':
            const_nq = Variable(
                th.Tensor([num_data * dim_latent]).type(float_type))
            # eqn (3.28) below p57 Damianou's Diss.
            KLD = 0.5 * (self.Xmean.pow(2).sum() + self.Xcov.transform().sum() -
                         self.Xcov.transform().log().sum() - const_nq)

        elbo -= KLD
        return -elbo

    def optimize(self, method='LBFGS', max_iter=2000, verbose=True, lr=0.01):
        """
        Optimizes the model by minimizing the loss (from :method:) w.r.t.
        model parameters.

        Args:
            method (torch.optim.Optimizer, optional): Optimizer in PyTorch
                (maybe add scipy optimizer in the future), default is `Adam`.
            max_iter (int): Max iterations, default 2000.
            verbose (bool, optional): Shows more details on optimization
                process if True.

        Todo:
            Add stochastic optimization, such as mini-batch.

        Returns:
            (np.array, value):

                losses: losses over optimization steps, (max_iter, )
                time: time taken approximately

        """
        parameters = ifilter(lambda p: p.requires_grad, self.parameters())

        if method == 'SGD':
            self.optimizer = th.optim.SGD(parameters, lr=0.05, momentum=0.9)
        elif method == 'Adam':
            self.optimizer = th.optim.Adam(parameters, lr=lr,
                                           betas=(0.9, 0.999), eps=1e-06, weight_decay=0.00001)
        elif method == 'LBFGS':
            self.optimizer = th.optim.LBFGS(parameters, lr=1, max_iter=5,
                                            max_eval=None, tolerance_grad=1e-05,
                                            tolerance_change=1e-09, history_size=50,
                                            line_search_fn=None)
        elif method == 'Adadelta':
            self.optimizer = th.optim.Adadelta(
                parameters, lr=1.0, rho=0.9, eps=1e-06, weight_decay=0.00001)
        elif method == 'Adagrad':
            self.optimizer = th.optim.Adagrad(
                parameters, lr=0.01, lr_decay=0, weight_decay=0)
        elif method == 'Adamax':
            self.optimizer = th.optim.Adamax(
                parameters, lr=0.002, betas=(0.9, 0.999), eps=1e-08, weight_decay=0)
        elif method == 'ASGD':
            self.optimizer = th.optim.ASGD(
                parameters, lr=0.01, lambd=0.0001, alpha=0.75, t0=1000000.0, weight_decay=0)
        elif method == 'RMSprop':
            self.optimizer = th.optim.RMSprop(
                parameters, lr=lr, alpha=0.99, eps=1e-08, weight_decay=0.00, momentum=0.01, centered=False)
        elif method == 'Rprop':
            self.optimizer = th.optim.Rprop(
                parameters, lr=0.01, etas=(0.5, 1.2), step_sizes=(1e-06, 50))
        # scipy.optimize.minimize
        # suggest to use L-BFGS-B, BFGS
        elif method in ['CG', 'BFGS', 'Newton-CG', 'Nelder-Mead', 'Powell', 'L-BFGS-B',
                        'TNC', 'COBYLA', 'SLSQP', 'dogleg', 'trust-ncg']:
            print('Scipy.optimize.minimize...')
            return self._optimize_scipy(method=method, maxiter=max_iter, disp=verbose)

        else:
            raise Exception('Optimizer %s is not found. Please choose one of the'
                            'following optimizers supported in PyTorch:'
                            'Adadelt, Adagrad, Adam, Adamax, ASGD, LBFGS, '
                            'RMSprop, Rprop, SGD, LBFGS. Or the optimizers '
                            'supported scipy.optimize.minminze: BFGS, L-BFGS-B,'
                            'CG, Newton-CG, Nelder-Mead, Powell, TNC, COBYLA,'
                            'SLSQP, dogleg, trust-ncg, etc.' % method)

        losses = np.zeros(max_iter)
        tic = time()

        print('{}: Start optimization via {} in the {} mode'.format(
            self.__class__.__name__, method,
            'inference' if self.inference else 'training'))

        if not self.inference:
            compute_loss = self.compute_loss
        else:
            compute_loss = self._compute_loss_inference
        if verbose:
            if not method == 'LBFGS':
                for iter in range(max_iter):
                    self.optimizer.zero_grad()
                    # forward
                    loss = compute_loss()
                    # backward
                    loss.backward()
                    self.optimizer.step()
                    losses[iter] = loss.data.numpy()
                    print('Iter: %d\tLoss: %s' % (iter, loss.data.numpy()))
            else:
                for iter in range(max_iter):
                    def closure():
                        self.optimizer.zero_grad()
                        loss = compute_loss()
                        loss.backward()
                        return loss
                    loss = self.optimizer.step(closure)
                    losses[iter] = loss.data.numpy()
                    print('Iter: %d\tLoss: %s' % (iter, loss.data.numpy()))
        else:
            if not method == 'LBFGS':
                for iter in range(max_iter):
                    self.optimizer.zero_grad()
                    # forward
                    loss = compute_loss()
                    # backward
                    loss.backward()
                    self.optimizer.step()
                    losses[iter] = loss.data.numpy()
                    if iter % 10 == 0:
                        print('Iter: %d\tLoss: %s' % (iter, loss.data.numpy()))
            else:
                for iter in range(max_iter):
                    def closure():
                        self.optimizer.zero_grad()
                        loss = compute_loss()
                        loss.backward()
                        return loss
                    loss = self.optimizer.step(closure)
                    losses[iter] = loss.data.numpy()
                    if iter % 10 == 0:
                        print('Iter: %d\tLoss: %s' % (iter, loss.data.numpy()))

        t = time() - tic
        print('Optimization time taken: %s s' % t)
        print('Optimization method: %s' % str(self.optimizer))
        if len(losses) == max_iter:
            print('Optimization terminated by reaching the maximum iterations\n')
        else:
            print('Optimization terminated by getting below the tolerant error\n')
        return losses, t

    def project(self, observ_test, observed_dims=None):
        """Infers the latent input corresponding to the new observed data
        The test data can be partially observed.
        # TODO: Currently only the Gaussian approximations, and iid case.
        With recognition model, inference of latent would be faster.

        Args:
            observ_test (numpy.ndarray): Test observed data
            observed_dims (list or np.array, 1D): Observed dimensions of the
                partially observed test data. Must be provided for partially
                observed test case.

        Returns:
            mean and variance of the posterior of Gaussian approximations
        """
        # Set the modes to be the inference mode
        self.inference = True
        if observed_dims is None:
            # Fully observable data
            assert observ_test.shape[1] == self.Y.size(1), \
                "Test data dimension must equal to that of the training " \
                "data for the fully observed case, otherwise please " \
                "specify the observed dimensions using ``observed_dims``"
        else:
            assert isinstance(observed_dims, (basestring, np.ndarray)), \
                "Type of the list of observed dimensions should be list " \
                "or 1d np.array"
            self.observed_dims = Variable(th.LongTensor(observed_dims))
        assert isinstance(observ_test, np.ndarray), "Test data should be " \
                                                    "np.ndarray"
        if observ_test.ndim == 1:
            observ_test = observ_test[None, :]
        # Design choice: do not create a tiny inference model, but reuse the
        # trained model using another function for compute the loss.
        # Add new observation variables to the original class
        self.Y_test = Variable(th.Tensor(observ_test).type(float_type))

        # Freeze the trained parameters
        for param in self.parameters():
            param.requires_grad = False

        # initialize Xmean_test, Xcov_test by searching for the nearest
        # neighbour in the data space
        if observed_dims is None:
            Y_observed = self.Y
        else:
            Y_observed = self.Y.index_select(1, self.observed_dims)

        YYT = self.Y_test.mm(Y_observed.t())
        dist_matrix = -2 * YYT + self.Y_test.pow(2).sum(1).expand_as(YYT) + \
                      Y_observed.t().pow(2).sum(0).expand_as(YYT)
        _, argmin = dist_matrix.min(1)
        argmin = argmin.view(self.Y_test.size(0)).data
        self.Xmean_test = Param(self.Xmean.data[argmin])
        self.Xcov_test = Param(self.Xcov.transform().data[argmin],
                               requires_transform=True)
        print("GPLVM: Finish preparing the model for projection")
        self._pre_compute()
        print("GPLVM: Done with pre-computation. \nPlease optimize the model"
              " again to obtain the projected latent variables\n")
        # optimize the latent variables
        # Q: how to know the optimization converges? this is slow and painful
        # Thus the model is returned to user for optimization
        # model_project.optimize(method='LBFGS', max_iter=100, verbose=False)
        # return model_project.Xmean, model_project.Xcov
        # self.optimize(method='LBFGS', max_iter=100, verbose=True)
        # use the ``compute_loss_inference`` method during the optimization
        # return self.Xmean_test, self.Xcov_test

    def _predict(self, Xnew_mean, Xnew_var=None, diag=True):
        """Computes the mean and variance of latent function output
        corresponding to the new (uncertain) input

        The new input can be deterministic or uncertain (only Gaussian: mean and
        variance). Returns the predictions over all dimensions (extract the
        needed dimensions for imputation case after getting the returns)

        Args:
             Xnew_mean (np.ndarray): new latent input, it is the deterministic
                input if ``input_var`` is None, otherwise it is the mean of the
                latent posterior, size n_* x q
             Xnew_var (np.ndarray): variance (covariance) of latent posterior,
                iid case, still n_* x q (each row stores the diagonal of cov)

        Returns:
            (Variables): n_* x p, mean of the predicted latent output
            (Variables): covariance of the predicted latent output,
                n_* x p for the deterministic case (share the same covariance),
                or n_* x q x q for the uncertain Gaussian input, iid.

        """
        assert isinstance(Xnew_mean, np.ndarray) and \
               Xnew_mean.shape[1] == self.Xmean.size(1), \
            "Input_mean should be numpy.ndarary, and its column dims " \
            "should be same as the latent dimensions"
        Xnew_mean = Variable(th.Tensor(Xnew_mean).type(float_type),
                             volatile=True)

        num_inducing = self.Z.size(0)
        beta = 1. / self.likelihood.variance.transform()
        # Psi1, Psi2
        eKxz = self.kernel.eKxz(self.Z, self.Xmean, self.Xcov)
        eKzxKxz = self.kernel.eKzxKxz(self.Z, self.Xmean, self.Xcov)
        Kzs = self.kernel.K(self.Z, Xnew_mean)
        Kzz = self.kernel.K(self.Z) + self.jitter.expand(self.Z.size(0)).diag()
        L = cholesky(Kzz, flag='Lkz')
        A = trtrs(L, trtrs(L, eKzxKxz).t()) * beta.expand_as(L)
        B = A + Variable(th.eye(num_inducing).type(float_type))
        Lb = cholesky(B, flag='Lb')
        C = trtrs(L, Kzs)
        D = trtrs(Lb, C)

        if Xnew_var is None:
            # broadcast udpated
            mean = D.t().mm(trtrs(Lb, trtrs(L, eKxz.t().mm(self.Y)))) \
                   * beta.expand(Xnew_mean.size(0), self.Y.size(1))
            # return full covariance or only the diagonal
            if diag:
                # 1d tensor
                var = self.kernel.Kdiag(Xnew_mean) - C.pow(2).sum(0).squeeze() \
                      + D.pow(2).sum(0).squeeze()
            else:
                var = self.kernel.K(Xnew_mean) - C.t().mm(C) + D.t().mm(D)
        else:
            # uncertain input, assume Gaussian.
            assert isinstance(Xnew_var, np.ndarray) and \
                   Xnew_var.shape == Xnew_var.shape, \
                "Uncertain input, inconsistent variance size, " \
                "should be numpy ndarray"
            Xnew_var = Param(th.Tensor(Xnew_var).type(float_type))
            Xnew_var.requires_transform = True
            Xnew_var.volatile = True
            # s for star (new input), z for inducing input
            eKsz = self.kernel.eKxz(self.Z, Xnew_mean, Xnew_var)
            # list of n_* expectations w.r.t. each test datum
            eKzsKsz = self.kernel.eKzxKxz(self.Z, Xnew_mean, Xnew_var,
                                          sum=False)
            Im = Variable(th.eye(self.Z.size(0)).type(float_type))
            E = trtrs(Lb, trtrs(L, Im))
            EtE = E.t().mm(E)
            F = EtE.mm(eKxz.t().mm(self.Y)) \
                * beta.expand(self.Z.size(0), self.Y.size(1))
            mean = eKsz.mm(F)
            Linv = trtrs(L, Im)
            Sigma = Linv.t().mm(Linv) - EtE
            # n x m x m
            # eKzsKsz = eKzsKsz.cat(0).view(Xnew_mean.size(0), *self.Z.size())
            var = []
            if diag:
                ns = Xnew_mean.size(0)
                p = self.Y.size(1)
                # vectorization?
                for i in range(ns):
                    cov = (self.kernel.variance.transform() -
                           Sigma.mm(eKzsKsz[i]).trace()).expand(p, p) + \
                          F.t().mm(eKzsKsz[i] - eKsz[i, :].unsqueeze(0).t().
                                   mm(eKsz[i, :].unsqueeze(0))).mm(F)
                    var.append(cov)
            else:
                # full covariance case, leave for future
                print("multi-output case, future feature")
                var = None
                pass

        return mean, var

    def generate(self, num_samples):
        """Generate new samples from the generative model

        Gaussian mixture model is a good choice for the iid latent.

        .. Note::
            Enforce the posterior of latents to approach the specified prior
            of latents, then samples from the prior, propagates through the
            model.This is the method used in VAEs. But the samples are not
            that good, visually (MNIST).

        .. Note::
            Two ways of drawing samples are different:
            1. Drawing one sample at a time and repeat multiple times ('random')
            2. Drawing multiple samples at a time (smooth)
        """
        # generate new samples from the posterior distributions

    def reconstruct(self, observed_part, observed_dims):
        """Reconstruct the missing dimensions in the test data

        Args:
            observed_part (np.ndarray): Partially observed test data
            observed_dims (slice): indices for the observed dimensions

        Returns:
            missing means and variances of test data
        """
        # 1. optimize q(X_*) - similar to projection
        self.project(observed_part, observed_dims)
        self.optimize(method='LBFGS', max_iter=100)
        # 2. generation / predict
        mean, var = self._predict(self.Xmean_test, self.Xcov_test)
        missing_dims = th.LongTensor(np.setdiff1d(range(self.Y.size(1)),
                                                  self.observed_dims))
        return mean[:, missing_dims], var[:, missing_dims]

    def _forecast(self, time_interval):
        pass

    # def evaluate_(self, labels):
    #     """Evaluate the learned latent embedding using:
    #     **nearest neighbour error**: the number of training data whose closest
    #     neighbour in the latent space corresponds to a data point of a different
    #     class.
    #
    #     Args:
    #         labels(np.ndarray): labels of the data
    #     """
    #     pass

    # def _reparam_vargp(self, Xmean_bar, Lambda):
    #     """
    #     Reparameterization of the variational gaussian process approximation.
    #
    #     Reference:
    #         @article{Opper:2009,
    #           title = {The Variational Gaussian Approximation Revisited},
    #           author = {Opper, Manfred and Archambeau, Cedric},
    #           journal = {Neural Comput.},
    #           year = {2009},
    #           pages = {786--792},
    #         }
    #
    #     Args:
    #         Xmean_bar (Tensor or Variable): the reparameterized mean, n x q
    #         Lambda (Tensor or Variable): the reparameterized covariance, n x q
    #
    #     Return:
    #         Xmean (Tensor): the orignal mean
    #         Xcov (Tensor): the original covariance matrix
    #     """
    #
    #     Kx = self.kernel_x.K(np.array(xrange(self.Y.size(0)))[:, None])
    #     Lkx = cholesky(Kx)
    #     # Kx_inverse = inverse(Kx)
    #     Xmean = Kx.mm(Xmean_bar)
    #     Xcov = []
    #     # S = []
    #     Le = []
    #     In = Variable(th.eye(self.Y.size(0)).type(float_type))
    #     for j in xrange(Xmean_bar.size(1)):
    #         Ej = Lkx.t().mm(Lambda[:, j].diag()).mm(Lkx) + In
    #         Lej = cholesky(Ej)
    #         Lsj = trtrs(Lej, Lkx.t()).t()
    #         Sj = Lsj.mm(Lsj.t())
    #         # potential improvement by Cholesky decomp
    #         # Sj = inverse(self.lambda_[:, j].diag() + Kx_inverse)
    #         Xcov.append(Sj.diag().unsqueeze(1))
    #         # S.append(Sj)
    #         Le.append(Lej)
    #     Xcov = th.cat(Xcov, 1)
    #
    #     return Xmean, Xcov
